Array.prototype.max = function(howMuchGreaterBThanA) {
    var max = this[0];

    var fCompare;
    if (typeof howMuchGreaterBThanA === 'function') {
        fCompare = function(item) {
            if (howMuchGreaterBThanA(max, item) > 0) {
                max = item;
            }
        }
    }
    else {
        fCompare = function(item) {
            if (item > max) {
                max = item;
            }
        }
    }

    this.forEach(fCompare);
    return max;
};

var ChatStorage = function() {
    function log(string) {
        opera.postError('!@[ChatStorage] ' + string);
    };
    
    var GRIND = 0.1;
    var ID_SEP = '-';   // !in {'.', ','}
    var CREATED = 'received'; // or 'sent'
    
    var MAX_LENGTH = 10000;
    var OUT_OF_LIMIT_MSG = '*** Single message limit reached: 10KB ***'; 

    this.__me = opera.io.webserver.userName;
    this.__topDirName = 'messengerDB';
    
    this.__cache = {};
    this.__cacheCapacity = 100;
    this.__keys = [];
    this.__cacheLen = 0;
    
    this.__touches = {};
    
    this.__init__ = function() {
        var storage = opera.io.filesystem.mountSystemDirectory('storage');
        this.__db = storage.resolve(this.__topDirName);

        if (!this.__db.exists) {
            this.__db = storage.createDirectory(this.__topDirName);
        }
        
        this.__db.refresh();
        
        // parse touches
        for (var d = 0, dir; dir = this.__db[d]; ++d) {
        
            dir.refresh();
            
            var init = dir[0] ? dir[0].name : 0;
            var first = last = +init;
        
            // find first and last
            for (var f = 1, file; file = dir[f]; ++f) {
                filedate = +file.name
                first = filedate < first ? filedate : first;
                last = filedate > last ? filedate : last;
            }
            
            // remember
            this.__touches[dir.name] = {
                first : first,
                last : last
            }
        }
        
        this.__clearCache();

    };
    
    this.__lastTouch = function(friend, timestamp) {

        if (!(friend in this.__touches)) {
            this.__touches[friend] = {
                first : timestamp,
                last : timestamp
            }
        }
        
        var last = this.__touches[friend].last;
        this.__touches[friend].last = last < timestamp
            ? timestamp : last;
    };
    
    /*
     * return last message to/from every friend
     * in one sorted (by received) array
     */
    this.getLastConversations = function() {
        // why not use this.__touches for that
        var last = [];
        
        function lastFirst(it, em) {
            return +em.received - (+it.received);
        };
        
        for (friend in this.__touches) {
        
            var dir = this.__db.resolve(friend);
            var fileName = this.__getFullUnit(this.__touches[friend].last);
            
            var file = dir.resolve(fileName);

            var jsdata = this.__getJsonFromFile(file);
            
            var data = [];
            for (id in jsdata) {
                jsdata[id].id = id;
                data.push(jsdata[id]);
            }
            
            if (data.length) {
                // message are stored by last received (aren't they?);
                last.push(this.__statifyItem(data.max(lastFirst), friend));
            }
        }
        
        // last conv. first
        return last.sort(lastFirst);
    };
    
    this.save = function(item, type) {
        var now = (new Date).valueOf();
        
        now = now <= this.__lastModified ? this.__lastModified + GRIND : now;
        
        item.received = now;
        
        if (item.message[MAX_LENGTH-1]) {
            item.message = item.message.substr(0, MAX_LENGTH) + OUT_OF_LIMIT_MSG;
        }
    
        var id = this.__makeId(item);
        
        this.__saveInCache(id, item);
        
        var dirName = this.__getDirName(item);
        var directory = this.__db.resolve(dirName);
        if (!directory.exists) {
            directory = this.__db.createDirectory(dirName);
        }
        
        // save by 'received'
        var fileName = this.__getFileName(item);
                
        var file = directory.resolve(fileName);
        
        file.appendContent('"'+ id + '":' + JSON.stringify(item) + ',');

        if (type == 'pending' || type == 'unread') {
            var key = dirName+'-'+type;
            var ids = widget.preferenceForKey(key);
            ids = ids ? ids.split(',') : [];
            ids.push(id);
            widget.setPreferenceForKey(ids.join(','), key);
        }
        
        this.__lastTouch(dirName, item[CREATED]);
        
        this.__setLastModified(item[CREATED]);
        
        return id;
    };
    
    this.noUnreadFromFriend = function(friend) {
        // friend.toLowerCase();
        this.__noMore(friend, 'unread');
    };
    
    this.noPendingToFriend = function(friend) {
        this.__noMore(friend, 'pending');
    };

    this.__noMore = function(friend, type) {
        friend = friend.toLowerCase();
        widget.setPreferenceForKey('', friend+'-'+type);
    };
    
    this.getItemById = function(id) {
        var item = this.__getFromCache(id);
        
        if (!item) {
            var dirName = this.__getDirName(id);
            var directory = this.__db.resolve(dirName);
            
            if (directory.exists) {

                var fileName = this.__getFileName(id);
                var file = directory.resolve(fileName);

                archive = this.__getJsonFromFile(file);
                
                item = archive ? archive[id] : null;

            }
        }
        
        return item ? (item.id = id, this.__statifyItem(item, dirName)) : null;
    };
    
    this.__getJsonFromFile = function(file) {
        var json = null;
        if (file.exists) {
            var content = file.readContent();
            if (content) {
                var inx = content.lastIndexOf(',');

                if (!!~inx) {
                    content = content.substr(0, inx);
                }
                content = '{' + content + '}';

                json = JSON.parse(content);
            }
        }
        return json;
    };
    
    this.__statifyItem  = function(item, dirName) {
        dirName = dirName || this.__getDirName(item);
        if (this.__isOfType(item.id, 'pending', dirName)) {
            item.state = 'pending';
        }
        else if (this.__isOfType(item.id, 'unread', dirName)) {
            item.state = 'unread';
        }
        return item;
    };
    
    this.howManyPendingToFriend = function(friend) {
        return this.__getFriendIdsOfType(friend, 'pending').length;
    };
    
    this.howManyUnreadFromFriend = function(friend) {
        return this.__getFriendIdsOfType(friend, 'unread').length;
    };
    
    this.getFromFriendByType = function(friend, type) {
        var data = [];
        var ids = this.__getFriendIdsOfType(friend, type);
        ids.forEach(function(id){
            var item = this.getItemById(id);
            if (item) {
                data.push(item);
            }
        }, this);
        return data;
    };
    
    this.__getFriendIdsOfType = function(friend, type) {
        friend = friend.toLowerCase();
        
        var ids = widget.preferenceForKey(friend+'-'+type);

        return ids ? ids.split(',') : [];
    };
    
    this.__isOfType = function(id, type, dirName) {
        dirName = dirName || this.__getDirName(id);
        
        var ids = this.__getFriendIdsOfType(dirName, type);
        
        return !!~ids.indexOf(id);
    };
    
    this.removeFromPending = function(id, friend) {
        this.__deleteIdOfType(id, 'pending', friend);
    };
    
    this.__deleteIdOfType = function(id, type, friend) {
        friend = friend || this.__getDirName(id);
        
        var ids = this.__getFriendIdsOfType(friend, type);
        
        var inx = ids.indexOf(id);
        
        if (!!~inx) {
            ids.splice(inx, 1);
            
            widget.setPreferenceForKey(
                ids[0] ? ids : '',    // non-empty
                friend+'-'+type
            );
        }
    };
    
    this.__escapeDelimiter = function(s)
    {
        if (typeof s == 'string' && s.length > 0)
        {
            var re = new RegExp("\\" + ID_SEP, "g");
            return s.replace(re, ''+ID_SEP+ID_SEP);
        }
        else
        {
            return s;
        }
    }
    
    // best solution would be to break up entities
    // code is hacky because '-' was originally used as ID
    this.__extractDelimPart = function(id, partNumber)
    {
        var parts = id.split(ID_SEP); 
        var s = "";
        var ret = [];
        
        for (var i = 0; i < parts.length; i++)
        {
            if (parts[i] != '')
            {            
                s += parts[i];
                if (parts[i+1] != '')
                {
                    ret.push(s);
                    s="";
                }
            }
            else 
            {
                s += "-";
            }
        }
        
        return ret[partNumber];
    }
    
    this.getRecipientById = function(id) {
        return this.__extractDelimPart(id, 1);
    };
    
    this.getSenderById = function(id) {
        return this.__extractDelimPart(id, 0);
    };

    this.getTimeById = function(id) {
        return this.__extractDelimPart(id, 2);
    };
    
    this.deleteItem = function(id) {
        var item;
        
        this.__deleteFromCache(id);
            
        var dirName = this.__getDirName(id);
        var directory = this.__db.resolve(dirName);

        if (directory.exists) {
            var fileName = this.__getFileName(id);
            var file = directory.resolve(fileName);
            var data = this.__getJsonFromFile(file);
            if (data) {
                item = data[id];
                item.id = id;
                item.state = 'deleted';
                
                delete data[id];
                
                var empty = true;
                for (anything in data) {
                    empty = false;
                    break;
                }
                
                if (empty) {
                    directory.deleteFile(file);
                }
                else {
                    var content = JSON.stringify(data);
                    var inx = content.indexOf('{');
                    content = content.substring(inx+1);
                    
                    inx = content.lastIndexOf('}');
                    content = content.substring(0, inx);
                    content += ',';
                    
                    file.writeContent(content);
                }
            }
        }
        
        this.__deleteIdOfType(id, 'pending', dirName);
        this.__deleteIdOfType(id, 'unread', dirName);

        this.__setLastModified((new Date).getTime());
        
        return item;
    };

    /*
     * getting: from < found <= to
     */
    this.getDataSince = function(from, to) {
    
        // get as much from __cache as possible
        from = (from || 0) + GRIND; // not inclusive
        to = to || Number.POSITIVE_INFINITY;
        
        var cacheData = [];
        
        var time, item, min = to;

        for (id in this.__cache) {
            time = this.getTimeById(id);
            min = min < time ? min : time;
            
            if (time >= from && time <= to) {
                item = this.__getFromCache(id);
                cacheData.push(this.__statifyItem(item));
            }
        }
        
        var data;

        // are there some messages left?
        if (min > from) {
            var diskData = this.__getDataFromDisc(from, min);

            // remove min from cacheData (it's doubled in diskData)
            cacheData = cacheData.filter(function(item) {
                return item[CREATED] != min
            }, this);
            
            data = cacheData.concat(diskData);
        }
        else {
            data = cacheData;
        }

        return data;    //. sort ???
                
    };
    
    this.getDataByUser = function(friend, from, to, num) {

        from = from || 0;
        to = to || Number.POSITIVE_INFINITY;
        num = num || Number.POSITIVE_INFINITY;
        
        var fromUnit = this.__getFullUnit(from);
        var toUnit = this.__getFullUnit(to, true);
    
        var data = [];
    
        var directory;
        if (typeof friend === 'string') {
            directory = this.__db.resolve(friend);
            friend = friend.toLowerCase();
            
            if (!directory.exists) {
                return data;
            }
        }
        else if (friend instanceof File && friend.isDirectory) {
            directory = friend;
        }
        else {
            throw "Wrong arguments: expects File of String";
        }
        
        directory.refresh();    // !mand
        
        // get proper files
        var files = Array.prototype.filter.call(
            directory,
            function(file) {
                var fileUnit = +file.name;
                return (fileUnit >= fromUnit && fileUnit <= toUnit);
            },
            this
        );
        
        // get proper items from sorted files
        files = files.sort(function(fi, les){
            return +les.name - +fi.name;
        });
        
        for (var i = 0, file; file = files[i]; ++i) {
            var json = this.__getJsonFromFile(file);

            // the following could be done in one loop
            // (created >= from && created <= to)
            // but this is more efficient (i think)
            if (+file.name < from) {
                for (var id in json) {
                    if (+json[id][CREATED] >= from) {
                        json[id].id = id;
                        data.push(this.__statifyItem(json[id], directory.name));
                    }
                }
            }
            else if (+file.name > to) {
                for (var id in json) {
                    if (+json[id][CREATED] <= to) {
                        json[id].id = id;
                        data.push(this.__statifyItem(json[id], directory.name));
                    }
                }
            }
            else {
                for (var id in json) {
                    json[id].id = id;
                    data.push(this.__statifyItem(json[id], directory.name));
                }
            }
            
            // TODO: optimize me!
            if (data.length >= num) {
                data = data.sort(function(it, em) {
                    return +em[CREATED] - +it[CREATED];
                }).slice(0, num);
            }
        }

        return data;
    };
    
    this.__getDataFromDisc = function(from, to) {
        this.__db.refresh();
        
        var data = [];
        
        for (var d = 0, dir; dir = this.__db[d]; ++d) {
            data.push(this.getDataByUser(dir, from, to)); 
        }

        // join data's subarrays into one array and return
        return Array.prototype.concat.apply([],data);
    };
    
    this.getTouchOfFriend = function(friend, first) {
        var touch;
        var kind = first ? 'first' : 'last';
        
        friend = friend.toLowerCase();
        
        if (friend in this.__touches) {
            touch = this.__touches[friend][kind];
        }
        
        return touch;
    };
    
    this.clear = function() {
        this.__db.refresh();
        for (var i = 0, dir; dir = this.__db[i]; ++i) {
            this.__db.deleteDirectory(dir, true);
            
            this.noUnreadFromFriend(dir.name);
            this.noPendingToFriend(dir.name);
        }
        this.__clearCache();
    };
    
    this.__getFileName = function(ii) {
        if (typeof ii === 'object') {
            time = ii[CREATED];
        }
        else {
            time = this.getTimeById(ii);
        }
    
        return this.__getFullUnit(time);
    };
    
    this.__getFullUnit = function(timestamp, forward) {
        if (timestamp == Number.POSITIVE_INFINITY) {
            return timestamp;
        }
    
        timestamp = forward ? +timestamp + 1000*60*60 : +timestamp;
        timestamp >> 0;
        
        var fullUnit = new Date(timestamp);
        fullUnit.setMilliseconds(0);
        fullUnit.setSeconds(0);
        fullUnit.setMinutes(0);
        
        return fullUnit.valueOf();
    };
    
    this.__getDirName = function(ii) {
        var sender, recipient;
        
        if (typeof ii === 'object') {
            sender = ii.sender;
            recipient = ii.recipient;
        }
        else {
            sender = this.getSenderById(ii);
            recipient = this.getRecipientById(ii);
        }

        return sender == this.__me ? recipient : sender;
    };
    
    this.__makeId = function(item) {
        return '' + this.__escapeDelimiter(item.sender) + ID_SEP + this.__escapeDelimiter(item.recipient) + ID_SEP + item[CREATED];
    };
    
    this.__saveInCache = function(id, item) {
        this.__cache[id] = item;
        this.__keys.push(id);
        ++this.__cacheLen;
        
        if (this.__cacheLen > this.__cacheCapacity) {
            var remId = this.__keys.shift();
            delete this.__cache[remId];
            --this.__cacheLen;
        }
    };
    this.__deleteFromCache = function(id) {
        var inx = this.__keys.indexOf(id);
        if (!!~inx) {
            this.__keys.splice(inx,1);
            --this.__cacheLen;
        }
        delete this.__cache[id];
    };
    this.__clearCache = function() {
        this.__cache = {};
        this.__keys = [];
        this.__cacheLen = 0;
    };
    this.__getFromCache = function(id) {
        if (id in this.__cache) {
            var item = {};
            for (field in this.__cache[id]) {
                item[field] = this.__cache[id][field];
            }
            item.id = id;
            return item;
        }
        return null
    };
    
    this.deleteFriend = function(friend) {
        friend = friend.toLowerCase();
        
        var directory = this.__db.resolve(friend);
        if (directory.exists) {
            this.__db.deleteDirectory(directory, true);
        }
    };
    
    this.__lastModified = parseInt(widget.preferenceForKey('last_modi')) || 0;
    
    this.__defineGetter__('lastModified', function() {
        return +this.__lastModified;
    });
    
    this.__setLastModified = function(timestamp) {
        this.__lastModified = +timestamp;
        widget.setPreferenceForKey(timestamp,'last_modi');
    };
    
    this.__init__();
};