/**
 *
 * This Friends Class has following goals:
 * 1. to keep a contact list
 * 2. to abstract unite users from devices (API 4 friend2friend communication)
 * 3. to keep track of the users getting on and off -line
 * 4. providing security means for friend2friend communication
 * 
 * @requires File.readContent, File.writeContent, File.appendContent
 *
 */
 
Array.prototype.indexByProperty = function(val, prop) {
    var len = this.length >>> 0;
    for (var i = 0; i < len; ++i) {
        if (this[i] && this[i][prop] === val) {return i;}
    }
    return -1;
};
 
 
function log() {
    opera.postError(arguments);
}

/**
 *  an abstract object for storing key-value pairs,
 *  as it is yet not decided whether quickStore object,
 *  or webstorage.js library will be used here
 */
var quickStore = new function() {
    this.setPreferenceForKey = function (pref, key) {
        widget.setPreferenceForKey(pref, key);
    };
    
    this.preferenceForKey = function(key) {
        return widget.preferenceForKey(key);
    };
};
 
/**
* @constructor
* @param {Boolean} stickToMyOpera If true, the contact list is to
*     reflect myOpera friends
* @param {String} filename Name of file (in storage folder) when the
*     the info about friends will be saved. Optional.
* @param {String} discoveryAction By means of this action Friends will
*     connect (make request) to other services instances. Optional. Use
*     when you plan on having multiple separate contact lists.
*/

function Friends(options) {
    options = options || {};
    
    // discover friends?
    ['discover', 'asd'].forEach(function(way) {
        if (typeof options[way] === 'undefined') {
            options[way] = true;
        }
        else {
            options[way] = !!options[way];
        }
    });
    
    options.action = options.action || 'friend';
    options.filename = options.filename || 'friends.lst';
    
    // interval for asd discovery [sec]
    options.asdInterval = options.asdInterval || 120;
    
    // interval for ffd discovery [sec]
    options.ffdInterval = options.ffdInterval || 15;
    
    // interval for querying myopera [sec]
    options.moInterval = options.moInterval || 3600 * 1000;
    
    // send my stuff to other devices of mine automatically?
    if (typeof distributeOnSuccess === 'undefined') {
        options.distributeOnSuccess = true;
    }
    else {
        options.distributeOnSuccess =  !!options.distributeOnSuccess;
    }
    
    // use double power asd when myOpera?
    if (typeof options.fullASD === 'undefined') {
        options.fullASD = true;
    }
    else {
        options.fullASD = !!options.fullASD;
    }
    
    // stick to my opera? if yes - what kind of?
    this.__moType = '';
    this.__myOpera = options.myOpera;
    if (this.__myOpera) {
        if (!~['mutual', 'fans', 'requested'].indexOf(this.__myOpera)) {
            this.__moType = '&type='+this.__myOpera;
        }
    }
    else {this.__myOpera = false;}
    
    this.__discover = options.discover;
    this.__asd = options.asd;
    
    this.__discoveryAction = options.action;
        
    this.__filename = options.filename;
    
    this.__friendsFile = opera.io.filesystem
        .mountSystemDirectory('storage')
        .resolve(this.__filename);
        
    this.__ffdInterval = options.ffdInterval;
    this.__asdInterval = options.asdInterval;
    this.__moInterval = options.moInterval;
        
            
    this.__friendsFile;    // contains info about:
    this.__friends;        // saved in friendsFile
    
    this.__changeHandlers = [];
    this.__userActions = {
        '__ruthere' : [function(evt){evt.connection.response.close()}]
    };
    
    this.__distributeOnSuccess = options.distributeOnSuccess;
    
    this.__fullASD = options.fullASD;
    
    this.__init__();
};

Friends.prototype = {
    constructor : Friends,
 
    __STATUS : {
        ONLINE : 2,
        OFFLINE : 1,
        UNDEFINED : -1
    },
    
    /**
     * Check if some1 is a friend (is in contact list)
     * @param {String} nick Friend's name.
     * @return {String} nick (toLowerCased or null)
     */
    isFriend : function (nick) {
        nick = nick.toLowerCase();
        return nick != this.__me && this.__isFriend(nick)
    },
    
    /**
     * The same as {@link Friends#isFriend} but does not
     * 'lowerCase' the argument and it includes 'me'
     * @private
     */
    __isFriend : function(nick) {
        return (nick in this.__friends) ? nick : null;
    },
    
    /**
     * Check if friend is (supposedly) online.
     * @param {String} nick Friend's name.
     * @return {Boolean}
     */
    supposedOnline : function(nick) {
        var ffdOnline = false;
        if (nick = this.isFriend(nick)) {
            for (var device in this.__friends[nick].devices) {
                var deviceItem = this.__friends[nick].devices[device];
                if (deviceItem.FFD == this.__STATUS.ONLINE) {
                    ffdOnline = true;
                    break;
                }
            }
        }
        return ffdOnline;  // || asdOnline; // || or &&?
    },
    
    /**
     * Add friend to contact list. (Checks if they're online).
     * @param {String} nick Friend's name.
     * @param {String} device Friend's device. Optional.
     * @return {Boolean}
     */
    addFriend : function(nick, device) {
        //proper url string
        if (nick !== encodeURIComponent(nick)) {return false;}

        var userName = nick;
        nick = this.__setFriend(nick.toLowerCase(), device);
        
        this.__setUserName(nick, userName);
        
        if (nick && device) {
////////
device = device.split('=')[0];
////////
            if (this.__discover) {
                this.__makeLetmeinRequest(nick, device);
            }
        }
    },
    
    
    /**
     * If device is defined, remove just the device;
     * remove (all info about) the friend otherwise.
     * Dispatches change event (removed if removed, offline
     * if deleting the device changed status to offline)
     * @param {String} nick Friend's name.
     * @param {String} device Friend's device. Optional.
     */
    removeFriend : function(nick, device) {
        
        if (!(nick = this.isFriend(nick))) {return;}
        
        var isThere = true;
        var before = this.supposedOnline(nick);
        
        if (device) {
            if (!(device in this.__friends[nick].devices)) {return;}
        
            delete this.__friends[nick].devices[device];
            
            quickStore.setPreferenceForKey('', nick+'-'+device+'-'+this.__filename);
            
            var after = this.supposedOnline(nick);
        }
        else {
            for (device in this.__friends[nick].devices) {
                quickStore.setPreferenceForKey('', nick+'-'+device+'-'+this.__filename);
            }
        
            delete this.__friends[nick];
            isThere = false;
        }
        
        if (!isThere) {
            this.__dispatchChangeEvent({removed : [this.__getFriend(nick)]});
        } else if (after != before) {
            this.__dispatchChangeEvent({offline : [this.__getFriend(nick)]});
        }
        
        this.__save();
    },
    
    /**
     * Return array of friends
     * @returns {Array} friends username )(id) and name
     */
    getFriends : function() {
        var friends = [];
        for (var nick in this.__friends) {
            //if (nick != this.__me) {nicks.push(nick);}
            var friend = this.__friends[nick];
            
            if (nick != this.__me) {
                friends.push(this.__getFriend(nick));
            }
        }
        return friends;
    },
    
    __getFriend : function(username) {
        //username = username.toLowerCase();
        friend = this.__friends[username];
        return {
            username : username,
            name : friend.fullName || friend.userName || nick
        };
    },
    
    /**
     * Get array of friend's devices
     * @param {String} nick Friend('s name) we want devices for.
     * @returns {Array} friend's devices(' names)
     */
    getDevicesFromFriend : function(nick) {
        var devices = [];

        if (nick = this.isFriend(nick)) {
            for (var device in this.__friends[nick].devices) {
                devices.push(device);
            }
        }
        return devices;
    },
    
    /**
     * Like {link #getDevicesFromFriend} but includes me
     * @private
     */
    __getDevicesFromFriend : function(nick) {
        var devices = [];

        if (this.__isFriend(nick)) {
            for (var device in this.__friends[nick].devices) {
                devices.push(device);
            }
        }
        return devices;
    },
    
    /**
     * Give the info of who's online and who's offline.
     * @returns {Object} about all friends online status
     * {
     *  online : [lolly, ichi, aga],
     *  offline : [hiruma, nina]
     * }
     */
    getStatus : function() {
        var online = [],
            offline = [];
            
        var friends = this.getFriends();
        friends.forEach(function(friend){
            if (this.supposedOnline(friend.username)) {
                online.push(friend)
            }
            else {
                offline.push(friend)
            }
        }, this);

        return {
            online : online,
            offline : offline,
            fans : +(this.__friends[this.__me].fans || 0)
        };
    },
    
    toString : function() {return '[Friends Object]';},
    
    /**
     * Sends an XmlHttpRequest to ALL devices of the friend.
     * @param {String} action Request's path (START with &, not ?, to be changed in future).
     * @param {String} recipient Friend to whom the request is made
     * @param {Object} stuff. Optional. Itself and all its fields, which could be:
     *  body : requests body
     *  headers : request headers as JSON
     *  onaccepted : function invoked ONCE iff at least one friend's device
     *      successfully responded to the request
     *  ondone : function invoked after all devices were sent request to
     *  self : object to use as this for onaccept and ondone
     */
    makeFriendRequest : function (action, recipient, stuff) {
        
        stuff = stuff || {};
        var that = stuff.self || stuff;
    
        if (!(recipient = this.__isFriend(recipient))) {
            // throw recipient + " is not on your contact list. Use addFriend() to add."
            if (typeof stuff.ondone === 'function') {
                stuff.ondone.call(that);
            }
            return recipient + " is not on your contact list. Use addFriend() to add."
        }

        var beforeStatus = this.supposedOnline(recipient);

        //var self = this;
        var originalOndone = stuff.ondone;
        var ondone = function() {
            if (beforeStatus !== this.supposedOnline(recipient)) {
                var event = {
                    online : [],
                    offline : []    // , removed : []
                }
                if (beforeStatus) {
                    event.offline.push(this.__getFriend(recipient));
                }
                else {
                    event.online.push(this.__getFriend(recipient));
                }
                this.__dispatchChangeEvent(event);
            }
            if (typeof originalOndone === 'function') {
                originalOndone.call(that);
            }
        };
        
        //shallowcopy
        var ext = {},
            adds = {};
        for (field in stuff) {
            ext[field] = stuff[field];
            adds[field] = stuff[field];
        }
        ext.self = this;
        adds.self = this;
        delete ext.ondone;
        delete ext.onaccepted;
        adds.ondone = ondone;
        adds.onaccepted = function() {
        
            if (this.__distributeOnSuccess && recipient != this.__me) {
                this.__makeGeneralRequest(action, this.__me, ext);
            }
            
            if (typeof stuff.onaccepted === 'function') {
                stuff.onaccepted.call(that)
            }
        };
        
        this.__makeGeneralRequest(action, recipient, adds);

    },
    
    /**
     * Add a handler for an action (made by makeFriendRequest). Some 
     * actions are not allowed. Adding them throws an error.
     * @param {String} action Used as action param of makeFriendRequest
     * @param {Function} handler This function is given an WebserverRequestEvent
     * parameter. Its connection field has additional parameter:
     * evt.connection.who = {sender : sendersNick, device : sendersDevice}.
     * The handler is itself responsible for dealing with the response.
     * The validation (if the request really comes from a friend) is done
     * beforehand.
     */
    addRequestListener : function(action, handler) {
        if (action in this.__reservedActions || action in this.__predefinedActions) {
            throw "I'm sorry, '" + action +"' is a reserved action of "
                + this.toString(); 
        }
        
        if (!(action in this.__userActions)) {
            this.__userActions[action] = [handler];
        }
        else if (!~this.__userActions[action].indexOf(handler)) {
            this.__userActions[action].push(handler);
        }
    },
    
    /**
     * Remove handler for an action (made by makeFriendRequest). Some 
     * actions are not allowed. Removing them throws an error.
     * @param {String} action Action to have handler removed from
     * @param {Function} handler Function to be removed
     */
    removeRequestListener : function(action, handler) {
        if (action in this.__reservedActions || action in this.__predefinedActions) {
            throw "I'm sorry, '" + action +"' is a reserved action of "
                + this.toString(); 
        }
        
        if (action in this.__userActions) {
            var inx = this.__userActions[action].indexOf(handler);
            if (!!~inx) {
                this.__userActions[action].splice(inx, 1);  //>?
            }
        }
    },
    
    /**
     * Add function for handling change event. That happens if
     * a status of some (or more friends) changed.
     * @param {Function} handler Function invoked when change
     * happens. It gets the parameter describing the change:
     *  {
     *      online : Array of nicks, that got online,
     *      offline : Array of nick, that got offline,
     *      removed : Array of nick, that gor removed
     *  }
     *  The above object may but may not contain these fields
     *  if they're empty. But will have at least one non-empty.
     */
    addChangeListener : function(handler) {
        this.__changeHandlers.push(handler);
    },
    
    /**
     * Remove function for handling change event. 
     * @param {Function} handler Function to be removed.
     */
    removeChangeListener : function(handler) {
        var inx = this.__changeHandlers.indexOf(handler);
        if (!!~inx) {
            this.__userActions[action].splice(inx, 1);  //>?
        }
    },
    
    /**
     * Calls handlers added with {link Friends#addChangeListener}
     * with difference Object as a parameter.
     * @private
     * @see #__checkFroDifference
     */
    __dispatchChangeEvent : function(event) {
        this.__changeHandlers.forEach(function(handler){
            handler(event);
        })
    },

    /**
     * compares results of getStatus() and dispatches the difference
     * @param {Object} before Object containing getStatus() result
     * @param {Object} after Object containing getStatus() result
     * @returns {Object} the difference
     * @private
     */
    __checkForDifference : function(before, after) {
        // is online afterwards, but weren't before
        var removed = before.online.concat(before.offline).filter(function(friend){
            return !~after.online.indexByProperty(friend.username, 'username') 
                && !~after.offline.indexByProperty(friend.username, 'username')
        });
        
        var gotOnline = after.online.filter(function(friend){
            return !~before.online.indexByProperty(friend.username, 'username');
        });
        
        var gotOffline = after.offline.filter(function(friend){
            return !~before.offline.indexByProperty(friend.username, 'username');
        });

        var difference = (
            gotOnline.length 
            || gotOffline.length 
            || removed.length
            || (before.fans != after.fans)
            ) ? {
                online : gotOnline,
                offline : gotOffline,
                removed : removed,
                fans : after.fans
            } : null;

        if (difference) {this.__dispatchChangeEvent(difference);}
        
        return difference;
    },
    
    /*
     * add friends if neccessery
     * add deviece if neccessery
     * define ONLINE/OFFLINE status (defaults to UNDEFINED)...
     * ... for method (defaults to FFD)
     * __save only if need to
     */
    __setFriend : function(nick, device, status, method) {
        var changed = false;
        // add the friend if need to
        if (!(this.__isFriend(nick))) {
            this.__friends[nick] = {devices:{}};
            changed = true;
        }
////////
if (device) {
    var arr = device.split('=');
    device = arr[0];
    var ipport = arr[1];
    var ip, port, path;
    if (ipport) {
        arr = ipport.split(':');
        ip = arr[0];
        var portpath = arr[1];
        if (portpath) {
            arr = portpath.split('>');
            port = arr[0];
            path = arr[1];
        }
    }
}
////////
        // add the device (if defined)
        if (device) {
            if (!(device in this.__friends[nick].devices)) {
            
                // proper URL
                if (device !== encodeURIComponent(device)) {return null;}
            
                this.__friends[nick].devices[device] = {
                    ASD : this.__STATUS.UNDEFINED,
                    FFD : this.__STATUS.UNDEFINED,
                    MO : this.__STATUS.UNDEFINED
                };
                changed = true;
            }
////////
if (ipport) {
    this.__friends[nick].devices[device].ip = ip;
    this.__friends[nick].devices[device].port = port;
    this.__friends[nick].devices[device].path = path ? '/'+path+'/' : undefined;
}
////////
            // is the status [& method] WELL defined?
            if (status in this.__STATUS) {
                method = method || 'FFD';
                method = this.__checkMethod(method, '__setFriend(' + arguments + ')');
                this.__friends[nick].devices[device][method] = this.__STATUS[status];
            }
        }
        
        // save if need to
        if (changed) {this.__save();}
        
        return nick;
    },
    
    /*
     * Check if it's ineed a valid discovery method
     */
    __checkMethod : function(method, where) {
        if (!~['FFD', 'ASD', 'MO'].indexOf(method)) {
            var error = this.toString() + ' unrecognised method: ' + method;
            if (where) {
                error += ' in ' + where;
            }
            throw error;
        }
        return method;
    },
    

    // DISCOVER

    /*
     * Discover all friends from the list (define their status)
     * using ASD. Dispatch (if) difference.
     */
    __ASD : function(callback, that) {

        var before = this.getStatus();
        
        var ondone = function(){

            var after = this.getStatus();
            
            this.__checkForDifference(before, after);

            if (typeof callback === 'function') {
                callback.call(that);
            }

        };
        
        if (this.__fullASD) {
            this.__discoverAllFriends('ASD', ondone, this);
        }
        else {  // __MO will discover my friends; I need to discover myself
            this.__discoverFriend('ASD', this.__me, ondone, this);
        }
    },
    
    /*
     * Discover all friends from the list (define their status)
     * using FFD (pinging). If ffdMethod == FFDletmein the
     * pinging is done with __letmein method; __ruthere otherwise.
     * Dispatch (if) difference.
     */
    __FFD : function() {
        arguments.callee.inProgress = true;
        var before = this.getStatus();

        this.__discoverAllFriends('FFD', function(){
        
            var after = this.getStatus();

            this.__checkForDifference(before, after);
            this.__FFD.inProgress = false;
            
            if (!this.__FFD.awake) {
                this.__FFD.awake = this.__FFD.awake === false;
            }

        }, this);
    },
    /* temp : 0, // debug */
    /*
     * Get all myOpera friends. Change the contact list, so
     * that is contains only my opera friends (see __redefineFriends).
     * Discovers all online friends. Dispatches the difference.
     */
    __MO : function(callback, that) {
        callback = (typeof callback === 'function') ? callback : function(){};
        var before = this.getStatus();
        sendRequest({   
            path : 'http://my.opera.com/community/api/users/friends.pl?user=' 
                + this.__Me + this.__moType,
            onsuccess : function(xhr) {
                var xml = xhr.responseXML;
                
                var myOperaFriends = xml.getElementsByTagName('friend');
                if (!myOperaFriends) {
                    myOperaFriends = [];
                } else {
                    myOperaFriends = Array.prototype.map.call(myOperaFriends, function(friend){
                        var fullName = friend.getAttribute("name");
                        var o = {userName : friend.textContent};
                        if (fullName) {o.fullName = fullName;}
                        return o;
                    });
                }
                
                myOperaFriends.push({userName : this.__Me});
                
                this.__redefineFriends(myOperaFriends, xml.documentElement.getAttribute('fans'));
                
                function dispatchDifference() {
                    var after = this.getStatus();
                    var difference = this.__checkForDifference(before, after);

                    var saved = false;
                    if (difference && difference.removed.length) {
                        this.__save();
                        saved = true;
                    }
                    callback.call(that);
                    // difference but was not saved?
                    return !!difference && !saved;
                };

                if (this.__discover) {
                    this.__discoverMyOperaFriends(dispatchDifference, this);
                    //__discoverMyOperaFriends does saving for us...
                }
                else {
                    dispatchDifference.call(this);
                }
            },
            onerror : function() {
                log('Could not connect to MO...');
                callback.call(that);
            },
            keep : 8,   // MO sometimes keeps the connection like for too long...
            self : this
        });
    },
    
    /*
     * this will ADD friends from users if neccessary
     * (keeps info about devices of friend's that remain)
     * delete friends that are NOT in the users array
     */
    __redefineFriends : function(users, fans) {
        fans = fans || 0;
        
        var modified = false;
        var friends = {};
        
        users.forEach(function(user){
            var userName = user.userName||'';
            var username = userName.toLowerCase();
            
            if (this.__isFriend(username)) {
                friends[username] = this.__friends[username];
                delete this.__friends[username];
            }
            else {
                modified = true;
                friends[username] = {devices:{}};
            }
            
            if (friends[username].fullName != user.fullName) {
                friends[username].fullName = user.fullName;
                modified = true;
            }
            if (friends[username].userName != user.userName) {
                friends[username].userName = user.userName;
                modified = true;
            }
        }, this);
        
        for (nomore in this.__friends) {
            modified = true;
            break;
        }
        
        if (friends[this.__me].fans != fans) {
            friends[this.__me].fans = fans;
            modified = true;
        }
        
        /******* SIMULATE FRIENDS *******/
        // this.oogabooga = false // !this.oogabooga;
        // if (this.oogabooga && this.__me == 'tommyandfriends') {
            // for (var i = 1; i <= 10; ++i) {
                // var usr = 'lolly' + i;
                // if (!(usr in friends)) {
                    // friends[usr] = {devices:{}};
                    // friends[usr].devices['device'] = {
                        // ASD : this.__STATUS.UNDEFINED,
                        // FFD : this.__STATUS.UNDEFINED,
                        // MO : this.__STATUS.UNDEFINED
                    // }
                // }
            // }
            // modified = true;
        // }
        /********************************/
        
        this.__friends = friends;
                
        if (modified) {this.__save();}
    },
    
    __discoverMyOperaFriends : function(callback, that) {
        //if (this.__me == 'iyondaime') this.__Me = 'iYondaime'
        sendRequest({
            path : 'http://asd.opera.com/1.0/users/' + this.__Me + '/friends/services',
            onsuccess : function(xhr) {

                var xml = xhr.responseXML;

                var users = this.__parseUsersDevices(xml);
                for (nick in this.__friends) {
                    for (device in this.__friends[nick].devices) {
                        this.__setFriend(nick, device, 'OFFLINE', 'MO');
                    }
                }

                for (username in users) {
                    users[username].forEach(function(device){       //
                        if (this.__setFriend(username, device.name, 'ONLINE', 'MO')) {
                            // if device's valid remember the path
                            this.__setPath(username, device.name, device.path);
                        }
                    }, this);
                }
            },
            onerror : function() {
                for (nick in this.__friends) {
                    for (device in this.__friends[nick].devices) {
                        this.__setFriend(nick, device, 'UNDEFINED', 'MO');
                    }
                }
            },
            ondone : function() {callback.call(that);},
            keep : 3,    // after keep secs the conn is assumed to have died
            self : this
        });
    },
    
    __parseUsersDevices : function(asdXml) {
        var services = asdXml.getElementsByTagName('service');
        services = services || [];
        
        var users = {};
        
        Array.prototype.forEach.call(services, function(service) {
            var path, device, username, originurl;

            if (
                (originurl = service.getAttribute('originurl'))
                && (originurl == this.__originURL)
                && (path = '/' + service.getAttribute('path') + '/')
                && (device = service.getAttribute('device'))
                && (username = service.getAttribute('username'))
            ) {
                username = username.toLowerCase();
                if (!(username in users)) {
                    users[username] = [];
                }

                users[username].push({
                    name : device,
                    path : path
                });
            }
        }, this);
        
        return users;
    },
    /*
     *
     */
    __discoverAllFriends : function(method, callback, that) {
        
        var call = typeof callback === 'function';
            
        var nicks = [];  // nicks are lowerCased
        for (nick in this.__friends) {nicks.push(nick)};

        if (!nicks.length) {
            if (call) {
                callback.call(that);
            }            
            return;
        }
        
        var checked = nicks.map(function(){return false});
        
        nicks.forEach(function(nick) {

            this.__discoverFriend(method, nick, function(success) {
                checked[nicks.indexOf(nick)] = true;
                if (checked.every(function(f){return f;})){
                    if (call) {callback.call(that);}
                }
            });
            
        }, this);
    },
    
    __discoverFriend : function(method, nick, callback, that) {
        // nick is lowerCased
        switch (method) {
            case 'ASD':
                
                if (typeof callback !== 'function') {
                    callback = function(){};
                }

                if (!(this.__isFriend(nick))) {
                    callback.call(that);
                    return;
                }

                var userName = this.__friends[nick].userName || nick;
                sendRequest({
                    path : 'http://asd.opera.com/1.0/users/' + userName + '/services',
                    onsuccess : function(xhr) {
                        var xml = xhr.responseXML;

                        var users = this.__parseUsersDevices(xml);
                        var devices = nick in users ? users[nick] : [];
                        
                        // set all to this.__STATUS.OFFLINE
                        for (device in this.__friends[nick].devices) {
                            this.__setFriend(nick, device, 'OFFLINE', 'ASD');
                        }

                        // set newly discovered to this.__STATUS.ONLINE
                        devices.forEach(function(device) {      //
                            if (this.__setFriend(nick, device.name, 'ONLINE', 'ASD')) {
                                this.__setPath(nick, device.name, device.path);
                            }
                        }, this);
                    },
                    onerror : function() {
                        for (device in this.__friends[nick].devices) {
                            this.__setFriend(nick, device, 'UNDEFINED', 'ASD');
                        }                        
                    },
                    ondone : function() {callback.call(that);},
                    keep : 3,    // after keep secs the conn is assumed to have died
                    self : this
                });
                break;
            case 'FFD':
                this.__makeGeneralRequest('__ruthere', nick, {ondone : callback, self : that});
                break;
            default:
                throw '__discoverFriend() - undefined discovery method: ' + method
                break;
        }
    },
    
    __hasJustPinged : function(nick, device, offline) {
        var dev;
        if (
            (this.__isFriend(nick)) 
            && (dev = this.__friends[nick].devices[device])
        ) {
            dev.lastPinged = (new Date).valueOf();
            if (!offline) {
                // last online
                quickStore.setPreferenceForKey(dev.lastPinged, nick+'-'+device+'-'+this.__filename);
            }
        }
    },
    
    __init__ : function() {

        var friendsTxt = this.__friendsFile.readContent();

        if (friendsTxt) {

            var friends = JSON.parse(friendsTxt);
            
            var lastInstance = +quickStore.preferenceForKey(this.__filename) || (new Date).getTime();
            var tooOld = 30 * 24 * 3600 * 1000;
            var saveNow = false;
            
            for (nick in friends) {
                var friend = friends[nick];
                
                //# denotes unneccessary stuff: it could as well NOT be saved
                //#! not neccessairly saved, but SHOULD be set to undefined/0

                // new instance: new keycodes to us
                // (but we'll keep 'toAccesskeys'
                delete friend['forAccesskey'];  //#
                delete friend['knockOn'];       //#

                // we don't know friends' online status
                for (device in friend.devices) {
                    friend.devices[device].ASD = this.__STATUS.UNDEFINED;   //#!
                    friend.devices[device].MO = this.__STATUS.UNDEFINED;    //#!
                    friend.devices[device].FFD = this.__STATUS.UNDEFINED;   //#!
                    friend.devices[device]['lastPinged'] = 0;               //#!
                    
                    var lastOnline = +quickStore.preferenceForKey(nick+'-'+device+'-'+this.__filename);
                    if (
                        lastInstance && lastOnline 
                        && lastInstance - lastOnline >= tooOld 
                    ) {
                        delete friend.devices[device];
                        quickStore.setPreferenceForKey('', nick+'-'+device+'-'+this.__filename);
                        saveNow = true;
                    }
                }
            }
            
            this.__friends = friends;
        }
        else {
            this.__friends = {};
        }
        
        this.__setFriend(this.__me);    // m3
        
        quickStore.setPreferenceForKey((new Date).getTime(), this.__filename);
        if (saveNow) {this.__save()};
        
        var self = this;
        opera.io.webserver.addEventListener(this.__discoveryAction, function(evt) {
            self.__discoverActionHandler(evt);
        }, false);
        
        this.connect();
    },
    
    disconnect : function() {
        clearInterval(this['connect'].interval)
        clearInterval(this['__startASD'].interval)
        clearInterval(this['__startFFD'].interval)
    },
    
    connect : function(later) {
        this.disconnect();

        if (!this.__asd) {
            this.__startFFD();
            return;
        }
        
        if (this.__myOpera) {
            this.__MO(this.__startASD, this);
            var self = this;
            arguments.callee.interval = setInterval(function(){
                self.__MO();
            }, this.__moInterval * 1000);
        }
        else {this.__startASD()};
    },
    
    __startFFD : function() {
        clearInterval(arguments.callee.interval);
        if (this.__discover) {
            this.__FFD();
            var self = this;
            arguments.callee.interval = setInterval(function(){
                self.__FFD();
            }, this.__ffdInterval * 1000);
        }
    },
    
    __startASD : function() {
        clearInterval(arguments.callee.interval);
        if (this.__discover) {
            this.__ASD(this.__startFFD, this);
            var self = this;
            arguments.callee.interval = setInterval(function(){
                self.__ASD();
            }, this.__asdInterval * 1000);
        }
    },
    
    __NOKEY : 'poo',    // a keycode generated by __createKeyCode should
                        // never be equal to __NOKEY
    
    /*
     * the __get* functions are to return either a valid keycode
     * or __NOKEY. (NEVER undefined)
     */
    __getAccesskeyTo : function(nick, device) {
        var dev;
        if (
            (this.__isFriend(nick)) 
            && (dev = this.__friends[nick].devices[device])
        ) {
            return dev.toAccesskey || this.__NOKEY;
        }
        return this.__NOKEY;
    },
    __setAccesskeyTo : function(accesskey, nick, device){
        var dev;
        if (
            (this.__isFriend(nick))
            && (dev = this.__friends[nick].devices[device])
        ) {
            return dev.toAccesskey = accesskey;            
        }
        return this.__NOKEY;
    },
    __assignAccesskeyTo : function(nick) {
        if (this.__isFriend(nick)) {
            return this.__friends[nick].forAccesskey = this.__createKeyCode();
        }
        return this.__NOKEY;
    },
    __getAccesskeyAssignedTo : function(nick) {
        if (this.__isFriend(nick)) {
            return this.__friends[nick].forAccesskey || this.__NOKEY;
        }
        return this.__NOKEY;
    },
    __createKeyCode : function() {
        return 'abcdefghjklmnopqrstuvwxyz0123456789'.split('')
            .sort(function(){return Math.random()-.5}).slice(0,12).join('');
    },
    __setKnockOn : function(nick) {
        if (this.__isFriend(nick)) {
            return this.__friends[nick].knockOn = this.__createKeyCode();
        }
        return this.__NOKEY;
    },
    __getKnockOn : function(nick) {
        if (this.__isFriend(nick)) {
            return this.__friends[nick].knockOn || this.__NOKEY;
        }
        return this.__NOKEY;
    },

    __validateRequest : function(request) {
        var accesskey = request.getRequestHeader('X-Accesskey');
        accesskey = accesskey ? accesskey[0] : null;
        
        var who = request.getRequestHeader('X-Device-User');
        who = who ? who[0] : null;

        if (who === encodeURI(who)) {   // '/path/'
            var arr = who.toLowerCase().split('.');
            if (arr.length >= 3) {
                who = {
                    sender : arr[1],
                    device : arr[0],
                    path : arr[2]
                };
            }
            else {who = null;}
        }
        else {who = null;}

        // get customAction from uri:
        //      /servicePath/discoveryAction/customAction?i=v&i2=v2
        //      /servicePath/discoveryAction/customAction/custom2/.../?i=v...
        var action = request.uri.split('/')[3];
        action = action ? action.split('?')[0] : null;
        
        var authorised = !!who && !!action && (
                accesskey
                && accesskey != this.__NOKEY
                && accesskey == this.__getAccesskeyAssignedTo(who.sender)
            );
        
        return {
            who : who,
            authorised : authorised,
            action : action
        };

    },
    
    __setPath : function (username, device, path) {
        var current = this.__friends[username].devices[device].path;
        if (path != current) {
            this.__friends[username].devices[device].path = path;
            this.__save();
        }
    },
    
    __setUserName : function (username, userName) {
        var current = this.__friends[username].userName;
        if (userName != current) {
            this.__friends[username].userName = userName;
            this.__save();
        }
    },
    
    __discoverActionHandler : function(evt) {
        var request = evt.connection.request,
            response = evt.connection.response;

        var validation = this.__validateRequest(request);

        if (!(validation.who && validation.action)) {
            // don't know who you are, bad request!
            response.setStatusCode(400);    // Bad Request
            response.close();
            return;
        }
        
        evt.connection.who = validation.who;
        
        var action = validation.action;

        if (!!~this.__reservedActions.indexOf(action)) {
            evt.connection.authorised = validation.authorised;
            this[action](evt);
        }
        else if (action in this.__userActions) {

            if (validation.authorised)
            {
                var setAccesskey = request.getRequestHeader('X-Set-accesskey');
                setAccesskey = setAccesskey ? setAccesskey[0] : null;
                
                // Q: What about __setFriend(sender, device) here?
                // A: Should not be neccessary as we must have made 
                // the '__welcome' request to that device before, 
                // which means we already know it
                
                // setAccesskey in this version is set only in '__welcome'
                // request, but just in case...
                if (setAccesskey) {
                    this.__setAccesskeyTo(
                        setAccesskey, 
                        validation.who.sender, 
                        validation.who.device
                    );
                }
                
                // our true friend requested
                this.__hasJustPinged(
                    validation.who.sender, validation.who.device
                );

                //@
                this.__setPath(validation.who.sender, validation.who.device, validation.who.path);
                
                this.__userActions[action].forEach(function(handler){
                    handler(evt);
                });
            }
            else {            
                // access denied
                response.setStatusCode(399);    // 401 Unauthorised
                response.close();
            }
        }
        else {
            // undefined action, (not found)
            response.setStatusCode(404);    // Not Found
            response.close();
        }
        
    },
        
    __reservedActions : ['__welcome', '__letmein'],
    __predefinedActions : ['__ruthere'],

    __welcome : function(evt) {

        var request = evt.connection.request,
            response = evt.connection.response, 
            who = evt.connection.who,
            authorised = evt.connection.authorised;
            
        var setAccesskey = request.getRequestHeader('X-Set-accesskey');
        setAccesskey = setAccesskey ? setAccesskey[0] : null;
            
        var knock = request.getItem('knock');
        knock = knock ? knock[0] : null;

        if (!(
            knock && knock != this.__NOKEY 
            && setAccesskey && setAccesskey != this.__NOKEY
        )) {
            // bad welcome request
            response.setStatusCode(400);    // Bad Request
            response.close();
            return;
        }
        
        // it's indeed 'welcome' to my 'letmein'
        if (knock == this.__getKnockOn(who.sender)) {
            //@
            this.__setPath(who.sender, who.device, who.path);
                    
            // state changed?
            // this must be BEFORE setFriend(...,ONLINE) of course ;]
            if (
                !this.__FFD.inProgress 
                && !this.supposedOnline(who.sender)                     //>
                && who.sender != this.__me
            ) {
                    this.__dispatchChangeEvent({online : [this.__getFriend(who.sender)]});//>
            }
            
            this.__setFriend(who.sender, who.device, 'ONLINE', 'FFD');  //> //

            this.__hasJustPinged(who.sender, who.device);
            
            // this shall be AFTER setFriend() in case it's a new device
            // (__setAccesskey can't add devices). This cannot happen
            // now, however, as __welcome request comes from devices
            // we sent __letmein to (those devices we therefore know)
            this.__setAccesskeyTo(setAccesskey, who.sender, who.device);
            
        } else {
            // someone else made the welcome request rather than the friend
            // i sent letmein to; IGNORE it.
        }
        
        //we're welcome so let's inform them if they have access to us
        if (!authorised) {
            response.setStatusCode(399);    // 401 Unauthorised
        }
        response.close();
        return;
    },
    
    __letmein : function(evt) {
        var request = evt.connection.request,
            response = evt.connection.response, 
            who = evt.connection.who;
            
        var knock = request.getItem('knock');
        knock = knock ? knock[0] : null;
        
        if (!(knock && knock != this.__NOKEY)) {
            // bad letmein request
            response.setStatusCode(400);    // Bad Request
            response.close();
            return;
        }
        
        // we have not yet __welcomed the friend; send them
        // status code that is counted as OFFLINE but is
        // different that the one used for Unauthorised
        response.setStatusCode(202);
        response.close();
        this.__makeWelcomeRequest(knock, who.sender, who.device, who.path);
    },
    
    __makeWelcomeRequest : function(knock, recipient, device, temppath) {
        // there could be a settable preference like: 
        // "only friends can send you messeges" or "(don't) allow
        // ppl who are not on your list to send you messeges"
        // FOR NOW don't welcome those you don't count friends
        if (this.__isFriend(recipient)) {   // recipient's lowerCased
        
            this.__setFriend(recipient, device);
            
            // if we want to welcome friend we must make grant them proper accesskey
            if (this.__getAccesskeyAssignedTo(recipient) === this.__NOKEY) {
                this.__assignAccesskeyTo(recipient);
            }
            
            this.__friends[recipient].devices[device].temppath = temppath;
            
            this.__makeSpecificDiscoveryRequest('__welcome?knock=' + knock, recipient, device);
        }
        // else: here is a place for letsbefriends request ;]
    },
    
    __makeLetmeinRequest : function(recipient, device, quiet) {
        // if we made a letmein request before, its better not to change
        // the 'knock' keyCode (it could make friend's __welcome requests
        // ivalid)
        var knock = this.__getKnockOn(recipient);
        knock = knock != this.__NOKEY ? knock : this.__setKnockOn(recipient);

        if (quiet) {
            this.__makeSpecificRequest(
                '__letmein?knock=' + knock, recipient, device, quiet
            );
        } else {
            this.__makeSpecificDiscoveryRequest(
                '__letmein?knock=' + knock, recipient, device
            );
        }
    },
    
    /*
     * remember recipient's online status
     * call __makeSepecificRequest, (onaccepted, body = null)
     * in ondone compare recipient's new online status and 
     * if different from previous one - dispatchChangeEvent
     */
    __makeSpecificDiscoveryRequest : function(action, recipient, device) {
        var beforeStatus = this.supposedOnline(recipient);
        var self = this;
        this.__makeSpecificRequest(
            action, recipient, device,
            {
                ondone : function() {
                    if (beforeStatus !== this.supposedOnline(recipient)) {
                        var event = {
                            online : [],
                            offline : []
                        }
                        if (beforeStatus) {
                            event.offline.push(this.__getFriend(recipient));
                        }
                        else {
                            event.online.push(this.__getFriend(recipient));
                        }
                        this.__dispatchChangeEvent(event);
                    }
                },
                headers : {'X-Set-accesskey' : this.__getAccesskeyAssignedTo(recipient)},
                self : this
            }
        );
    },
    
    /*
     * make the 'action' request to given device.recipient
     * call (just ONCE) onaccepted if any device is online (and accepted the request)
     * call ondone when done
     * set devices FFD online status accordingly 
     * @param requestData = { action : m [,body : xxx[, headers : {}]]}
     */
    __makeSpecificRequest : function(action, recipient, device, stuff) {
    
        stuff = stuff || {};
        var that = stuff.self || stuff;
        
        // don't send stuff to myself...
        if (recipient == this.__me && device == this.__deviceName) {
            if (typeof stuff.ondone === 'function') {
                stuff.ondone.call(that);
            }
            return;
        }

        var usePath = action.split('?')[0] === '__welcome' ?
            'temppath' : 'path';
            
        var servicePath = this.__friends[recipient].devices[device][usePath]
            || this.__servicePath;
            
////////
if (this.__friends[recipient] && this.__friends[recipient].devices[device]) {
    var ip = this.__friends[recipient].devices[device].ip;
    var port = this.__friends[recipient].devices[device].port;
}
var path = 'http://';
path += (ip ? ip : device + '.' + recipient + '.'
            + this.__proxyName);
path += port ? ':' + port : '';
path += servicePath + this.__discoveryAction;
////////
        
        var headers = {
            'X-Device-User' : this.__deviceName + '.' + this.__me + '.' + this.__servicePath,
            'X-Accesskey' : this.__getAccesskeyTo(recipient, device),
            'Content-Type' : 'text/plain',
        };

        for (header in stuff.headers) {
            headers[header] = stuff.headers[header]
        }
        
        sendRequest({
            path : path + '/' + action,
            headers : headers,
            body : stuff.body,
            onsuccess : function() {
                this.__setFriend(recipient, device, 'ONLINE', 'FFD');
                
                if (usePath === 'temppath') {   // response to __welcome...
                    this.__setPath(
                        recipient, 
                        device,
                        this.__friends[recipient].devices[device][usePath]
                    )
                }

                if (typeof stuff.onaccepted === 'function') {
                    stuff.onaccepted.call(that);
                }
            },
            '202' : function() {
                // can't set 'ONLINE' (friend's __welcome req is to do that).
                // Also, can't set 'OFFLINE' coz friend's __welcome
                // will reach us faster than this response (strange :|)
            },
            onerror : function() {
                this.__setFriend(recipient, device, 'OFFLINE', 'FFD');
            },
            '399' : function() {    // Unauthorised
                // if they're online but we can't communicate, it's like off
                this.__setFriend(recipient, device, 'OFFLINE', 'FFD');
                
                if (usePath === 'temppath') {   // response to __welcome...
                    this.__setPath(
                        recipient, 
                        device,
                        this.__friends[recipient].devices[device][usePath]
                    )
                }

                // against inf loops (just in case)
                if (action === '__letmein') {
                    opera.postError('This should NOT happen: '
                        + '__letmein handler is never to set 399 status');
                    return;
                }
                this.__makeLetmeinRequest(recipient, device);
            },
            ondone : function() {
                if (typeof stuff.ondone === 'function') {
                    stuff.ondone.call(that);
                }
                //log(arguments[0].status, path + '/' + action);
            },
            self : this,
            keep : 5
        });
        
        this.__hasJustPinged(recipient, device);
    },
    
    /*
     * make the 'action' request to all recipient's devices
     * call onaccepted if device accepted the request
     * call ondone after checking all devices
     */
    __makeGeneralRequest : function (action, recipient, stuff) {
        
        // if we want to assign a new accesskey for a friend
        // this is a place to do it
        // if (i want to assign new acceskey for some reason) {
        //    this.__assignAccesskeyTo(recipient);
        //}

        var devices = this.__getDevicesFromFriend(recipient);
        var devicesLeft = devices.length + 1;

        var stuff = stuff || {};
        var that = stuff.self || stuff;
        
        function checkIfDone() {
            --devicesLeft;
            if (!devicesLeft) {
                if (typeof stuff.ondone === 'function') {
                    stuff.ondone.call(that);
                }
            }
        };
        checkIfDone();
        
        var accepted = false;
        function onAccepted() {    // accepted
            if (!accepted) {
                accepted = true;
                if (typeof stuff.onaccepted === 'function') {
                    stuff.onaccepted.call(that);
                }
            }
            checkIfDone();
        }
        
        var adds = {};  
        // shallo copy
        for (field in stuff) {adds[field] = stuff[field]};
        adds.ondone = checkIfDone;
        adds.onaccepted = onAccepted;
        
        if (action === '__ruthere') {
            
            var now = (new Date).getTime();
            
            devices.forEach(function(device) {

                var deviceItem = this.__friends[recipient].devices[device];
                
                // how much time has gone since we last communicated?
                var timeApart = now - (deviceItem.lastPinged || 0);
                
                if (deviceItem.FFD == this.__STATUS.ONLINE) {
                    if (timeApart > this.__ffdInterval * 1000) {
                        this.__makeSpecificRequest(
                            '__ruthere', recipient, device, adds
                        );
                    }
                    else {checkIfDone();}
                }
                else {  // !ONLINE
                    if (!this.__FFD.awake || timeApart > this.__ffdInterval * 20 * 1000) {
                        // make 'quiet' __letmein
                        this.__makeLetmeinRequest(recipient, device, adds);
                    }
                    else {checkIfDone();}
                }
            
            }, this);
            
        } // it's not a pinging request
        else {
            
            devices.forEach(function(device) {
                if (
                    this.__friends[recipient].devices[device].FFD 
                    == this.__STATUS.ONLINE
                ) {
                    this.__makeSpecificRequest(action, recipient, device, adds);
                }
                else {checkIfDone();}
                
            }, this);
            
        }
    },
    
    /*
     * saves data about friends
     */
    __save : function() {
        return this.__friendsFile.writeContent(JSON.stringify(this.__friends));
    },
    
    __servicePath : opera.io.webserver.currentServicePath,
    __proxyName : opera.io.webserver.proxyName,
    __deviceName : opera.io.webserver.deviceName,
    __Me : opera.io.webserver.userName,
    __me : opera.io.webserver.userName.toLowerCase(),
    __originURL : opera.io.webserver.originURL
    
};  // eof Friends.prototype