/**
 * Constructs a new Notifier object.
 *
 * @class Notifier provides a method for showing notifications (it uses
 * widget.showNotification()). 
 * Notification that happen whithin a predefined interval are handled together.
 * ATM Opera shows only one line of notification messages; all the content may 
 * be viewed in error console on callback.
 * @constructor
 * @return A new notifier
 */
function Notifier(defaults)
{
    defaults = defaults || {};
    var INTERVAL = defaults.interval || 5000;
    
    var messages = [];
    
    var notifying = false;
            
    /** Queues a notification
     * @param msg {String} The message to queue
     * @param callback {Function} The callback to be executed when notification is executed
     */
    function notify(msg, callback)
    {
        if (msg) {messages.push(msg);}

        if (messages.length) {

            if (notifying) {return;}

            notifying = true;
            
	                
	    if (! callback)
	    {
		callback = defaults.callback;
	    }
	    
            // wait a mo: maybe some other notification will be pushed,
            // then show them all together
            
            setTimeout(function() {
                msg = messages.map( function(m)
                                    {
                                       try
                                       {
                                           return decodeURIComponent(m);
                                       }
                                       catch (e)
                                       {
                                           return m;
                                       }
                                    }
                                ).join(' \n'); // \n is ignored for now (a bug) so I'm adding a whitespace
        messages = [];
	    widget.getAttention();
	    widget.showNotification(msg, callback);
	    notifying = false;
	    notify();
            }, INTERVAL);
            
        }
    };

    this.notify = notify;
};