/** Helper for setting correct localization for datejs using Accept-Language header from browser */
var DateJsLocalizer = new function()
{
    var _defaultLanguage        = "en-US.js",
        _cachedLanguages        = "",
        _cachedReturnLanguage   = "";
    
    var findLocal = function(languages)
    {
        if (languages)
        {
            var langs = languages.split(",");
            for (var i = 0; i < langs.length; i++)
            {
                langs[i] = langs[i].split(";")[0];
            }
        }
        
        var locals = opera.io.filesystem.mountSystemDirectory("application").resolve("public_html/globalization");
        locals.refresh();
        for (var i = 0; i < langs.length; i++)
        {
            for (var j = 0; j < locals.length; j++)
            {
                if (locals[j].isFile && locals[j].name.indexOf(langs[i]) > -1)
                {
                    return locals[j].name;
                }
            }
        }
        
        return "";
    }
    
    /**
    * Will find correct JavaScript file that can be included before DateJs for correct locale.
    * Will cache result and only seek for a new file if the languages parameter is changed.
    * Will search in public_html/globalization for correct file
    * @param languages The contents of the Accept-Language header delivered from the browser
    * @returns A string containing the name of the file to include.
    */ 
    this.localize = function(languages)
    {
        var langs = [];
        var returnLang = _defaultLanguage;
        
        if (_cachedLanguages == languages)
        {
            // no need to look, we can return the cached value
            returnLang = _cachedReturnLanguage;
        }
        else
        {
            var lang = findLocal(languages);
            if (lang != "")
            {
                returnLang = lang;
            }
        }
        
        _cachedLanguages        = languages;
        _cachedReturnLanguage   = returnLang;

        return returnLang;
    }
}