/**
 * =]
 * @requires Registry, Friends, EventsManager, fileUtilities
 */
function PresenceChat() {

    function log(str) {
        opera.postError('PrCh^^^^   ' + str);
    }

    var translator = new Translator({domain: 'service'});
    
    
    // === DATA HANDLING === //
    
    var asd = true;

    // friends
    var myFriends = new Friends({
        myOpera : true, // stick to myopera friends
        asd : asd,      // use asd
        fullASD : false // 'friends/services' only
    });

    var chatStorage = new ChatStorage();

    // client communication event manager; handles events pushed to client
    var EventsManager = new CCEManager();
    
    
    
    // === HELPERS === //
    
    var notifier = new Notifier();
    
    /**
     * mark data's messages as 'read'
     */
    function readify(data) {
        data.forEach(function(item) {
            delete item.state;
        });
        return data;
    };
    
    function authorize(connection) {
        // change it to Yusef's validation?
        if (!connection.isOwner) {
            requestError(connection.response, 'Unauthorised');
            return false;
        }
        return true;
    };
    
    function buildFriendsEvent(evt) {
        body = {};
        var hasFriends = widget.preferenceForKey("seen_friends") || false;
        var seenFriends = false;
        
        ['offline','online'].forEach(function(status){
            if (evt[status] && evt[status].length) {
                body[status] = evt[status].map(function(user){
                    var username = user.username;

                    user.numunread = chatStorage.howManyUnreadFromFriend(username);
                    user.haspending = !!chatStorage.howManyPendingToFriend(username);
                    
                    var firsttime = chatStorage.getTouchOfFriend(username, true);
                    var lasttime = chatStorage.getTouchOfFriend(username, false);
                    
                    if (firsttime) {user.firsttime = firsttime;}
                    if (lasttime) {user.lasttime = lasttime;}
                    
                    seenFriends = true;
                    return user;
                });
            }
        });
        
        if (evt['removed']){body['removed'] = evt['removed']};
        if (evt['fans']){body['fans'] = evt['fans']};

        if (!hasFriends && seenFriends)
        {
            widget.setPreferenceForKey("1", "seen_friends");
        }
        
        return JSON.stringify(body);
    };
    
    var Statuses = {
        'Unauthorised' : '401',
        'Bad Request' : '400',
        'Conflict' : '409'
    };
    
    function requestError(response, kind) {
        var status = (kind in Statuses) ? Statuses[kind] : '400';
        response.setStatusCode(status);
        response.close();
    };
    
    // -- utilities --
    WebServerRequest.prototype.parseItems = function(items, firstOnly) {
        
        if (!(items instanceof Array)) { 
            items = [items];
        }
        
        var parsed = {
            success : true,
            items : {}
        };
        
        items.forEach(function(item){
            parsed.items[item] = null;
            
            // getItem won't work sometimes (build 1537); no idea why...
            var arr = this.getItem(item) || this.bodyItems[item];
            
            if (!arr) {
                parsed.success = false;
            }
            else {
                parsed.items[item] = firstOnly 
                    ? arr[0] 
                    : Array.prototype.map.call(arr, function(x){return x;});
            }
        }, this);
        
        return parsed;
    };    
    
    var webserver = opera.io.webserver;
    var me = webserver.userName;
    
    // === ACTIONS === //

    function acceptRequest(evt) {

        var request = evt.connection.request,
            response = evt.connection.response,
            who = evt.connection.who;

        if (request.body) {    // !imonline request
            // parse data
            var parsed = [];
            try {
                var posts = JSON.parse(request.body);
                posts.forEach(function(post){
                    if (
                        // recipient and sender are ok?
                        !(
                            (who.sender == me)
                            || (
                                post.recipient == me 
                                && post.sender == who.sender
                            )
                        ) 
                        
                        // the message is of string type?
                        || typeof post.message !== 'string'
                        
                        // post sent is string or number
                        || (
                            typeof post.sent !== 'string' 
                            && typeof post.sent !== 'number'
                        )
                    // if not:
                    ) {throw 'requestError';}
                    
                    // post may contain additional fields like: id, received
                    // we save only relevant data
                    parsed.push({
                        sender : post.sender,
                        recipient : post.recipient,
                        message : post.message,
                        sent : post.sent
                    });

                })
            }
            catch(e) {
                return requestError(response);
            }
            
            // save in history
            parsed.forEach(function(post) {
                chatStorage.save(post,'unread');
            });
            
            EventsManager.registerEvent('chat');    // accepted
            
            // notify
            if (!(clientFocused && clientAwake)) {
                notifier.notify(parsed[0].sender + ': ' + parsed[0].message,
                                function() {
                                    widget.setPreferenceForKey(parsed[0].sender, "currentuser");
                                });
            }
            // all tabs may be closed, but we're not yet in sleep mode...
            else if (!EventsManager.isOpen()) {
                window.setTimeout(function() {
                    // sure it's not open?
                    if (!EventsManager.isOpen()) {
                        notifier.notify(parsed[0].sender + ': ' + parsed[0].message,
                                        function() {
                                    widget.setPreferenceForKey(parsed[0].sender, "currentuser");
                                });

                    }
                }, 200);
            }
            
        } // eof if (request.body)
        response.close();

    };
    
    /*
     * checks if there's any pending stuff to a friend;
     * if there is, an attempt to send it is being made
     */
    function sendPendingToFriend(friend) {
        var username = friend.username;
        var pendingStuff = chatStorage.getFromFriendByType(username, 'pending');

        if (!pendingStuff.length) {return;}

        myFriends.makeFriendRequest(
            'accept',
            username, 
            {
                body : JSON.stringify(pendingStuff),
                onaccepted : function() {
                    chatStorage.noPendingToFriend(username);
                    EventsManager.registerEvent('chat', readify(pendingStuff));  //@
                }
            }
        );
    };
    
    // === SET LISTENERS FOR EVENTS PUSHED TO CLIENT === //
    
    EventsManager.setEventListener('friend', function(connection, data) {
        var response = connection.response;
        response.setResponseHeader('X-Action', 'friends');
        response.write(data);
        response.close();
        return true;
    });
    
    var __START = (new Date).getTime();

    EventsManager.setEventListener('chat', function(connection, data) {

        var request = connection.request,
            response = connection.response;
            
        var lastUpdated = request.getRequestHeader('X-Last-updated');
        lastUpdated = lastUpdated ? +lastUpdated[0] : __START;

        var responseData = data || chatStorage.getDataSince(lastUpdated);

        if (responseData.length) {

            response.setResponseHeader('X-Action', 'chat');
            
            if (!data) {
                response.setResponseHeader('X-Updating-to', chatStorage.lastModified);
            }

            response.write(JSON.stringify(responseData));
            response.close();
        }
        return true;
    });
    
    /*EventsManager.setEventListener('switchToFriend', function(connection, data) {
         var response = connection.response;
         response.setResponseHeader('X-Action', 'switchToFriend');
         response.write(data);
         response.close();
         return true;
     });
    */
    
    
    // === SET CLIENT-PUSHED EVENT TRIGGERS === //
    
    // handle accept requests from friends
    myFriends.addRequestListener('accept', acceptRequest);

    // handle changes in friends status
    myFriends.addChangeListener(function(evt){
        EventsManager.registerEvent('friend', buildFriendsEvent(evt));
        
        if (evt.online) {
            evt.online.forEach(sendPendingToFriend);
        }
    });
    
    
    // === (CLIENT'S)REQUESTS HANDLING === //
    
    var clientFocused = false;
    var clientAwake = true;
  
    /*
     * Handlers' methods provide means to handle requests
     * made by owner's UA
     */
    var Handlers = new function() {

        this.focusstate = function(evt){
            var request = evt.connection.request,
                response = evt.connection.response;
                
            var data = request.parseItems('focusstate', true);
            if(!data.success) {
                return requestError(response);
            }
            
            clientFocused = data.items.focusstate === 'true';
            
        };
    
        this.listen = function(evt) {

        };
        
        /*
         * save a message and send it to the recipient's service
         */
        this.send = function(evt) {
    
            var request = evt.connection.request,
                response = evt.connection.response;
            
            var data = request.parseItems(['recipient', 'message'],true);
            if (!data.success) {
                return requestError(response);
            }
            
            var recipient = data.items.recipient,
                message = data.items.message;
            
            var sender = me;
            var sent = (new Date).getTime();
            
            // save in pending
            var post = {
                sender : sender,
                recipient : recipient,
                message: message,
                sent : sent
            };

            var id = chatStorage.save(post, 'pending');   // save as pending
            var accepted = false;
            // send to the friend
            myFriends.makeFriendRequest('accept', recipient, 
                {
                    body : JSON.stringify([post]), 
                    onaccepted : function() {
                        // move from pending to read
                        chatStorage.removeFromPending(id, recipient);
                        EventsManager.registerEvent('chat', [chatStorage.getItemById(id)], true);    // send
                        accepted = true;
                    },
                    ondone : function() {
                        EventsManager.registerEvent('chat', false, !accepted);    // send
                    }
                }
            );
        };

        this.friendsshow = function (evt) {
            var online = [];
            var offline = [];
        
            evt.connection.response.setResponseHeader('X-Action', 'friends');
            
            var friends = myFriends.getStatus();
            
            evt.connection.response.write(buildFriendsEvent(friends));
            evt.connection.response.close();

        };
        this.friendsshow.priv = true;
        
        this.getmessages = function (evt) {
            var response = evt.connection.response,
                request = evt.connection.request;

            var data = request.parseItems(['user', 'from', 'to', 'num'], true);
            if (!data.items.user) {
                return requestError(response);
            }
            
            var responseData = chatStorage.getDataByUser(
                data.items.user, data.items.from, data.items.to, data.items.num
            );

            response.setResponseHeader('X-Action', 'chat');        //>?

            response.write(JSON.stringify(responseData));
            response.close();
        
        };
        this.getmessages.priv = true;
        
        this.deletemessages = function(evt) {
            var response = evt.connection.response,
                request = evt.connection.request;
                
            var data = request.parseItems('id');
            if (!data.success) {
                return requestError(response);
            }
            
            var responseData = [];

            data.items.id.forEach(function(id) {
                var item = chatStorage.deleteItem(id);
                if (item) {
                    responseData.push(item);
                }
            });
            
            EventsManager.registerEvent('chat', responseData, true);
        };
        
        this.readmessages = function(evt) {
            var response = evt.connection.response,
                request = evt.connection.request;

            var data = request.parseItems('user', true);
            if (!data.success) {
                return requestError(response);
            }

            EventsManager.registerEvent('chat', 
                readify(chatStorage.getFromFriendByType(data.items.user, 'unread')),
                true
            );
            chatStorage.noUnreadFromFriend(data.items.user);
        };
        
        this.getconversations = function(evt) {

            var response = evt.connection.response;
            
            response.setResponseHeader('X-Action', 'conversations');        //>?
            response.write(JSON.stringify(chatStorage.getLastConversations()));
            response.close();
        };
        this.getconversations.priv = true;
        
        this.currentuser = function(evt)
        {
            var request = evt.connection.request,
            response = evt.connection.response;
        
            if (request.method.toUpperCase() == "POST")
            {
                var bodyItems = request.bodyItems["user"];
                var user = "";
                if (bodyItems && bodyItems.length)
                {
                    user = bodyItems[0];
                }
                //opera.postError("setver set " + user);
                widget.setPreferenceForKey(user, "currentuser")
            }
            
            var user = widget.preferenceForKey("currentuser");
            response.setStatusCode(200);
            response.setResponseHeader('X-Action', 'currentuser');
            response.write('<user>');
            if (user)
            {
                //opera.postError("server return " + user);
                response.write(user);
            }
            response.write('</user>');
            response.close();
        }
        this.currentuser.priv = true;
    };
    
    for (handler in Handlers) {
        (function(action) {
            webserver.addEventListener(action, function(evt) {
                if (authorize(evt.connection)) {
                    if (!Handlers[action].priv) {
                        EventsManager.plug(evt.connection);
                    }
                    //log(action)
                    Handlers[action](evt);   
                }
            }, false);
        })(handler)
    };

    webserver.addEventListener('client.js', function(evt) {
        translator.sendClientScript( evt.connection ); 
    } , false);
    
    /*
     * _index handler
     */
    webserver.addEventListener('_index', function(evt) {
        var r = evt.connection.response;
        var connection = evt.connection;
        var template = '';
        var data =
        {
            servicePath :''+opera.io.webserver.currentServicePath,
            deviceName  :''+opera.io.webserver.deviceName,
            username    :''+opera.io.webserver.userName,
            serviceName :''+opera.io.webserver.currentServiceName,
            hostName    :''+opera.io.webserver.hostName,
            seenFriends :''+widget.preferenceForKey("seen_friends") ? true : false,
            loadTime    :''+(new Date()).getTime(),
            messengerURI:connection.request.protocol+'://'+connection.request.headers.Host[0].replace(/admin./, "")+connection.request.uri
        }
        
        var langs = (evt.connection.request.getRequestHeader('Accept-Language')||[])[0];
        data.dateJsCultureFile = DateJsLocalizer.localize(langs);
        translator.setLanguagesFromHttp(evt.connection);
        
        if (evt.connection.isOwner)
        {
            // normal chat interface
            template = 'templates/index.html';
        }
        else
        {
            // debug: force one of the templates to show, so we don't have to
            // load up different browsers to see them (remove all instances of 'view' later)
            var view = evt.connection.request.queryItems['view'];
            view = view && view[0];
            
            if (evt.connection.isLocal && view!='opera' && view !='unite' && view != 'other')
            {
                // page indicating that you have some messages, go to nromal interface
                
                template = 'templates/overview.html';
                // this needs to be done server side since if it is '1' it should
                // not be a plural "messages".
              
            }
            else
            {
                // sniffing the user agent here only displays a slightly
                // different message for different browsers, the options available
                // to them all are the same however.
                
                var userAgent = evt.connection.request.headers['User-Agent'];
                userAgent = (userAgent && userAgent[0] || '').toLowerCase();
                
                data.userAgent = userAgent;
                
                if (userAgent.indexOf('opera')+1 && view !='other' || view == 'opera' || view =='unite')
                {
                    var version = userAgent.match(/.*opera\/([0-9\.]+)\s.*/)
                    version = version && parseFloat(version[1]) || 0;
                    
                    if (version >= 9.8 && view != 'opera' || view == 'unite') // opera 10 or above, should be enough
                    {
                        template = 'templates/index_noaccess_unite.html'
                    }
                    else
                    {
                        template = 'templates/index_noaccess_opera.html'
                    }
                }
                else
                {
                    template = 'templates/index_noaccess.html'
                }
            }
        }
        
        var template = new Markuper(template, data);
        var string = translator.translate(template.parse().html(), langs);
        r.write(string);
        r.close();
    }, false);
    
    
    
    // === OTHER ;) (temporary) === //
    
    /* 
     * helpful commands for testing
     */ 
    /*=========================================================================
    webserver.addEventListener('clear', function(evt) {
        if (evt.connection.isOwner) {
            chatStorage.clear();
            evt.connection.response.write('cleared');
        }
        evt.connection.response.close();
    }, false);
    =========================================================================*/
    
    
    // === INIT! === //
    EventsManager.setModeChangeListener(function(awake) {
        clientAwake = awake;
        if (awake) {
            log('Wake up! connecting...');
            myFriends.connect();
        }
        else {
            log('Entering sleep mode. Disconnecting...')
            myFriends.disconnect();
        }
    })
    EventsManager.watchdog = true;
    EventsManager.registerEvent('chat');    //initial

};  // eof PresenceChat


// adding this:
new PresenceChat();
