
var MessageFilterCtrl = new function()
{
    this.setActiveFilter = function(filterId)
    {
        var els = $('message-filter-list').getElementsByTagName("li");
        for (var i = 0, el; el = els[i]; i++)
        {
            var re = new RegExp("^mf_" + filterId);
            if (el.id && el.id.match(re))
            {
                el.addClass("filter-selected");
            }
            else if (el.id && el.hasClass("filter-selected"))
            {
                el.removeClass("filter-selected");
            }
        }
    }
    
    this.onFilterClicked = function(evt)
    {
        var id = evt.target.id || evt.target.parentNode.id;

        if (id.match(/_(\d+)/))
        {
            if ((MessageData.getMessageFilter(ViewState.CurrentSelectedUser) != RegExp.$1))
            {
                // TODO: be more clever about clearing the view
                $("message-view-feed").innerHTML = ""; 
                MessageData.setMessageFilter(ViewState.CurrentSelectedUser, RegExp.$1);
                MessageFilterCtrl.setActiveFilter(RegExp.$1);
                MessageCtrl.showMessagesForCurrentUser();
            }
        }    
    }
}

var HeaderCtrl = new function()
{
    var lastFocused = null;
    
    var removeLastFocused = function()
    {
        if (lastFocused)
        {
            lastFocused.removeClass("active_header");
        }
    }
    
    this.focusConversation = function()
    {
        removeLastFocused();
        $("conversation_header").addClass("active_header");
        
        $("message").addClass("conversation");
        $("message").removeClass("help");
        lastFocused = $("conversation_header");
        ViewCtrl.enableForm(false, true);
    }
    
    this.focusUser = function()
    {
        removeLastFocused();
        $("username_header").addClass("active_header");
        
        $("message").removeClass("conversation");
        $("message").removeClass("help");
        lastFocused = $("username_header");    
        if (ViewState.CurrentSelectedUser)
        {
            ViewCtrl.enableForm(true);
        }
    }
    
    this.focusHelp = function()
    {
        removeLastFocused();
        $("help_header").addClass("active_header");
        
        $("message").removeClass("conversation");
        $("message").addClass("help");
        lastFocused = $("help_header");
    }
    
    this.onConversationClicked = function(evt)
    {
        MessageCtrl.onConversationSelected();
    }
    
    this.onUserClicked = function(evt)
    {
        MessageCtrl.onUserSelected(ViewState.CurrentSelectedUser);
    }
}

var ContactsCtrl = new function()
{
    var _numContacts = 0;
    
    this.getContactEl = function(username)
    {
        return $("contact_" + username);
    }
    
    this.hasContact = function(username)
    {
        return (ContactsCtrl.getContactEl(username) != null);
    }
    
    this.hasContacts = function()
    {
        return (_numContacts > 0);
    }
    
    this.removeContact = function(username)
    {
        var el = $("contact_" + username);
        el.parentNode.removeChild(el);
        if (0 == --_numContacts)
        {
            ViewCtrl.displayEmptyScreen();
        }
    }

    this.removeFan = function(username)
    {
        MessageData.setIsFan(username, false);
    }
    
    this.getContainerForContact = function(username, isOnline)
    {
        if (isOnline)
        {
            ret = $("friends-online");
        }
        else
        {
            ret = $("friends-offline");
        }
        
        return ret;
    }
    
    this.addAlpha = function(liEl, container)
    {
        var gotAdded = false;

        if (container.childNodes.length > 0)
        {
            for (var i = 0, el; el = container.childNodes[i]; i++)
            {
                var hasLessPriority = (parseInt(liEl._priority) < parseInt(el._priority));
                var hasHigherPriority = (parseInt(liEl._priority) > parseInt(el._priority));
                var isAlphabeticallyBefore = (liEl.textContent < el.textContent);
                
                if (hasHigherPriority || (isAlphabeticallyBefore && !hasLessPriority))
                {
                    el.parentNode.insertBefore(liEl, el);
                    gotAdded = true;
                    break;
                }
            }
        }
        
        if (! gotAdded) // or if it has no child nodes
        {
            container.appendChild(liEl);
        }
    }
    
    this.updateState = function(username)
    {
        var el = $("contact_" + username);

        if (el)
        {
            var isOnline = MessageData.isOnline(username);
            var numUnread = MessageData.getNumUnread(username);
            var hasPending = MessageData.hasPending(username);

            var unreadEl = el.getElementsByClassName("num-unread")[0];
            var pendingEl = el.getElementsByClassName("pending-messages")[0];
			
	    // This boolean tracks if we need to re-order the contact in the list of contacts
            var shouldForceUpdate = false; 
            
            if (numUnread > 0)
            {
                shouldForceUpdate = (el._priority != 1);
                el._priority = 1;
                
                if (numUnread > 99)
                {
                    numUnread = 99;
                }
                
                unreadEl.textContent = numUnread;
                unreadEl.style.display = "inline";
                pendingEl.style.display = "none"; // do not show pending if we have unreads
            }
            else
            {
                shouldForceUpdate = (el._priority != 0);
                el._priority = 0;
                unreadEl.style.display = "none";                
				
                if (hasPending)
                {
                    pendingEl.style.display = "inline";
                }
                else
                {
                    pendingEl.style.display = "none";
                }
            }
            
            if (isOnline)
            {
                this.setContactOnline(username);    
            }
            else 
            {
                this.setContactOffline(username, shouldForceUpdate);
            }
        }
    }
    
    var createContactLIEl = function(username, name, numUnread)
    {
        var liEl = document.createElement("li");
        var aEl = document.createElement("a");
        aEl.href = "#";
        
        var imgEl = document.createElement("img");
        imgEl.src = "http://my.opera.com/" + username + "/picture.pl?xscale=40";
        imgEl.addEventListener("error", function() { this.src = "img/blank_avatar60.png" }, false);
        liEl.appendChild(imgEl);
        
        var userNameEl = document.createElement("span");
        if (name == "" || name == undefined)
        {
            userNameEl.appendChild(document.createTextNode(username));
        }
        else
        {
            userNameEl.appendChild(document.createTextNode(name));
        }
        userNameEl.className = "username";
        liEl.appendChild(userNameEl);
        
        var numUnreadEl = document.createElement("span");
        numUnreadEl.className = "num-unread";
        numUnreadEl.textContent = numUnread;
        liEl.appendChild(numUnreadEl);
        
        var pendingEl = document.createElement("span");
        pendingEl.className = "pending-messages";
        liEl.appendChild(pendingEl);
        
        liEl.addEventListener("click", function() {}, false);
        
        return liEl;
    }
    
    /** (friend["username"], true, friend["numunread"], friend["haspending"]); numUnread, hasPending, isFan */
    this.addContact = function(username, isOnline, data)
    {
        if (! username)
        {
            throw "Username is null";
        }
        var isFan = false;
        var id = "contact_" + username;
        var liEl = $(id);
        
        if (!data)
        {
            data = {};
        }

        if (username === ViewState.CurrentSelectedUser && data.numunread) {
            var body = 'user=' + encodeURIComponent(ViewState.CurrentSelectedUser);
            ViewCtrl.connection.startAndKeep('readmessages', body);
            data.numunread = 0;
        }
        MessageData.setNumUnread(username, data.numunread);
        
        MessageData.setPending(username, data.haspending);
        MessageData.setIsOnline(username, isOnline);
        if (data.firsttime)
        {
            MessageData.setFirstMessageTime(username, data.firsttime);
        }
        
        _numContacts++;

        if (! liEl)
        {
            liEl = createContactLIEl(username, data["name"], data.numunread);
            liEl.id = id;
            
            // TODO: move to MessageData? 
            liEl._data =
            {
                isOnline : isOnline
            }
            
            if (data.numunread)
            {
                liEl._priority = 1;
            }
            else
            {
                liEl._priority = 0;
            }

            this.addAlpha(liEl, this.getContainerForContact(username, isOnline));
            
            if (! isOnline)
            {
                liEl.addClass("offline");
            }
            
            if (isFan)
            {
                liEl.addClass("fan");
            }
            
            _numContacts++;
        }
        else
        {
            this.updateState(username);
        }

        
        ViewCtrl.displayNormalScreen();
        ViewCtrl.updateFriendsMoreButton();
    }
    
    this.setContactOnAlert = function(username)
    {
        var el = ContactsCtrl.getContactEl(username);
        
        if (el.parentNode.id != "friends-on-alert")
        {
            el.parentNode.removeChild(el);
            this.addAlpha(el, $("friends-on-alert"));
        }
    }
    
    this.setContactOnline = function(username)
    {
        var el = ContactsCtrl.getContactEl(username);

        if (el && false == el._data.isOnline)
        {
            el._data.isOnline = true;
            el.removeClass("offline");
        }
        
        if (el.parentNode.id != "friends-online")
        {
            el.parentNode.removeChild(el);
            this.addAlpha(el, $("friends-online"));
        }
        
        ViewCtrl.updateFriendsMoreButton();
    }
    
    this.setContactOffline = function(username, force)
    {
        var el = ContactsCtrl.getContactEl(username);
  
        if (el && el._data.isOnline)
        {
            el._data.isOnline = false;
            el.addClass("offline");
        }
        
        if (el.parentNode.id != "friends-offline" || force)
        {
            el.parentNode.removeChild(el);
            this.addAlpha(el, $("friends-offline"));
        }
        
        ViewCtrl.updateFriendsMoreButton();
    }
        
    
    this.setContactSelected = function(username)
    {
        //username may be undefined/null. that's ok. will clear selected.
        var target = this.getContactEl(username);
        var lis = $("contacts").getElementsByTagName("li");
        for (var n=0, li; li=lis[n]; n++)
        {
            if (li==target)
            {
                li.addClass("selected");
                continue;
            }
            else
            {
                li.removeClass("selected");
            }
        }
        this.updateState(username);
    }
}

var MessageCtrl = new function()
{
    var isAtBottom = true; // used to track if the user has manually changed scroll position, if so, shouldn't autoscroll to bottom
    var messageViewEl = null;
    var unsentMessages = {}; // username is key and message is value
    var self = this;
    
        // smiley codes, support all variations that Opera irc chat client supports
    var SMILIES = {
        ':@':    'angry.png',
        ':-@':   'angry.png',
        'B-)':   'cool.png',
        '8-)':   'cool.png',
        ":'(":   'cry.png',
        ':D':    'lol.png',
        ':-D':   'lol.png',
        ':-|':   'normal.png',
        ':|':    'normal.png',
        "'<":    'pacman.png',
        ':(':    'sad.png',
        ':-(':   'sad.png',
        ':)':    'smile.png',
        ':-)':   'smile.png',
        ':o':   'surprised.png',
        ':O':   'surprised.png',
        ':-O':   'surprised.png',
        ':b':    'tongue.png',
        ':-b':   'tongue.png',
        ':p':    'tongue.png',
        ':-p':   'tongue.png',
        ':P' :   'tongue.png',
        ':-P' :  'tongue.png',
        ';)':    'wink.png',
        ';-)':   'wink.png'
    };
    
    
    this.clearAllMessages = function()
    {
        messageViewEl.innerHTML = "";
    }
    
    this.getMessageHeight = function(el)
    {
        var computedStyle = document.defaultView.getComputedStyle($("message-view"), null);
        return computedStyle.height.match(/(\d+)/)[1];
    }
    
    this.onScrollView = function(evt)
    {
        var elHeight = MessageCtrl.getMessageHeight();
        if ($("message-view").scrollTop >= ($("message-view").scrollHeight - elHeight))
        {
            isAtBottom = true;
        }
        else
        {
            isAtBottom = false;
        }
    }
    
    this.scrollIntoPosition = function()
    {
        if (isAtBottom)
        {
            var elHeight = MessageCtrl.getMessageHeight();
            $("message-view").scrollTop = $("message-view").scrollHeight - elHeight;
        }
    }
    
    this.storeUnsentMessage = function(username)
    {
        var unsentMessage = $("message_text").value;
        
        if (unsentMessage)
        {
            unsentMessages[username ? username : ViewState.CurrentSelectedUser] = unsentMessage;
        }
    }
    
    this.restoreUnsentMessage = function(username)
    {
        var unsentMessage = unsentMessages[username ? username : ViewState.CurrentSelectedUser];
        if (unsentMessage)
        {
            $("message_text").value = unsentMessage;
        }
        else
        {
            $("message_text").value = "";
        }
    }
    
    this.clearUnsentMessage = function(username)
    {
        unsentMessages[username ? username : ViewState.CurrentSelectedUser] = null;
    }
    
    this.onHelpClicked = function()
    {
        HeaderCtrl.focusHelp();
        ViewCtrl.enableForm(false);
        //ViewCtrl.onUserSelected("");
    }
    
    this.onDeletePendingClicked = function(evt)
    {
        var match = [];
        if (evt.target.parentNode.parentNode.id.match(/message_(.*)/))
        {
            var body = "id=" + RegExp.$1;
            ViewCtrl.connection.startAndKeep('deletemessages', body);
        }
    }
    
    this.deletePendingMessage = function(post)
    {
        var el = $("message_" + post.id);
        if (el)
        {
            el.parentNode.removeChild(el);
        }
        MessageData.deletePendingMessage(post);
        ContactsCtrl.updateState(ViewState.CurrentSelectedUser);
    }
    
    this.onUserSelected = function(username)
    {
        HeaderCtrl.focusUser();

        if (username && ViewState.CurrentSelectedUser != username)
        {
            // set current user on server side
            ViewCtrl.connection.startAndKeep('currentuser', "user=" + username);
            // reset message filter
            MessageData.setMessageFilter(username, MessageFilterSpan.Today);
            MessageFilterCtrl.setActiveFilter(MessageFilterSpan.Today);
        }

        $("remove_user").style.display = $("remove_fan").style.display = "inline";
        $("remove_user").textContent = "Remove " + username;
        $("remove_fan").textContent = "Unfan " + username;
        
        $("username_header").textContent = username;
        
        // store the unsent message for user before switching
        MessageCtrl.storeUnsentMessage();
        ViewState.CurrentSelectedUser = username;  
        MessageCtrl.clearAllMessages();
        // restore potential unsent message after switching user
        MessageCtrl.restoreUnsentMessage();
        
        if (! MessageData.hasMessages(username))
        {
            // has no messages for this user. Request messages. note that the response will
            // happen in an async manner and UI will only be updated when/if event is returned
            var body = 'user=' + encodeURIComponent(username);
            ViewCtrl.connection.startAndKeep('getmessages', body);
            
            // the startAndKeep should support queue so safe to do this at once
            body = 'user=' + encodeURIComponent(ViewState.CurrentSelectedUser);
            ViewCtrl.connection.startAndKeep('readmessages', body);
        }
        
        // add person's name and thumbnail to the info area
        infoEl = $('messenger_info');
        infoEl.innerHTML = ''; // clear
        infoEl.href = "http://my.opera.com/" + username + '/';
        
        imgEl = document.createElement('img');
        imgEl.src = "http://my.opera.com/" + username + "/picture.pl?xscale=100";
        imgEl.addEventListener("error", function() { this.src = "img/blank_avatar60.png"; }, false);
        imgEl.alt = username+'\'s avatar.';
        infoEl.appendChild(imgEl);
        
        $("message_text").focus();
        MessageCtrl.onMessageTextInput();

        MessageData.setNumUnread(username, 0);
        ContactsCtrl.setContactSelected(username);
        
        // reloading the page will preselect the same user
        window.location.hash = username;
    }
    
    /**
     * @returns {Element} A list item to go in the chat feed
     */
    this.createMessageHTML = function(msg)
    {
        var li = document.createElement('li');
        var parent = null;
        var child = null;
        var lines = msg.message.split(/\r\n|\r|\n/);
        var words = null;
        var i = 0;
        var url = /\b(((https?|ftp|irc|telnet|nntp|gopher|file):\/\/|(mailto|news|data|opera):)[^\s\"<>\{\}\'\(\)]*)/g;
                
        li.id = 'message_' + msg.id;
        
        parent = document.createElement('label');
        
        // add time
        child = document.createElement('span');
        child.textContent = new Date(msg.received).toString("HH:mm");
        parent.appendChild(child);
        
                // add name
        child = document.createElement('span');
        child.textContent = msg.sender;
        parent.appendChild(child);
        
        li.appendChild(parent);
        
        // add pending
        if ("pending" == msg.state)
        {
            li.setAttribute('title', _('Waiting for {user} to come online', { 'user' : msg.recipient }));
            
            parent = document.createElement('div');
            parent.className = 'pendingBox';
            
            child = document.createElement('img');
            child.src = 'img/anim_pending.gif';
            parent.appendChild(child);
            
            child = document.createElement('button');
            child.setAttribute('type', 'button');
            child.title = _('Cancel sending this message');
            child.addEventListener("click", MessageCtrl.onDeletePendingClicked, false);
            parent.appendChild(child);
            
            li.appendChild(parent);
        }
        
        // add message body
        parent = document.createElement('div');
        parent.className = 'message-text';
        
        for (i = 0; i < lines.length; i++)
        {
            words = lines[i].split(' ');
            for (j = 0; j<words.length; j++)
            {
                var word = words[j];
                
                // handle smilies
                if (word in SMILIES)
                {
                    words[j] = document.createElement('img');
                    
                    words[j].src = 'img/smilies/' + SMILIES[word];
                    words[j].alt = word;
                    words[j].title = word;
                    continue;
                }
                
                // handle urls
                if (word.match(url))
                {
                    words[j] = document.createElement('a');
                    words[j].href = word;
                    words[j].target = "_blank";
                    words[j].appendChild(document.createTextNode(word));
                    continue;
                }
                
                // otherwise add to previous string to avoid calling
                // createTextNode too many times
                if (j>0 && (typeof words[j-1] == 'string'))
                {
                     // add to previous word
                    words[j-1] += ' '+words[j];
                    
                    // remove current word
                    words.splice(j,1);
                        
                    // reduce counter since words array has changed
                    j--;
                    continue;
                }
            }
            
            child = document.createElement('p');
            for (j=0; j<words.length; j++)
            {
                if (typeof words[j] == 'string')
                {
                    child.appendChild(
                        document.createTextNode(
                            (j>0?' ':'') + // space before
                            words[j] +
                            (j<words.length-1?' ':'') // space after
                        )
                    );
                }
                else
                {
                    if (j>0 && (typeof words[j-1] != 'string'))
                    {
                        // make sure there is a space between smilies and links
                        child.appendChild(document.createTextNode(' '));
                    }
                    child.appendChild(words[j]);
                }
            }
            
            parent.appendChild(child);
        }
        
        li.appendChild(parent);
        
        if (Config.myUsername == msg.sender)
        {
            li.className = "ourself"; 
        }
        
        return li;
    }
    
    this.createConversatonHTML = function(msg)
    {
        var li = document.createElement('li');
        var parent = null;
        var child = null;
        var lines = msg.message.split(/\r\n|\r|\n/);
        var words = null;
        var i = 0;
        var url = /\b(((https?|ftp|irc|telnet|nntp|gopher|file):\/\/|(mailto|news|data|opera):)[^\s\"<>\{\}\'\(\)]*)/g;
                
        li.id = 'conversation_' + msg.id;
        
        parent = document.createElement('label');
        
        // add name
        child = document.createElement('span');
        child.textContent = (msg.sender == Config.myUsername) ? msg.recipient : msg.sender;
        parent.appendChild(child);
        
        // add time
        child = document.createElement('span');
        child.textContent = new Date(msg.received).oclock();
        parent.appendChild(child);
        
        
        li.appendChild(parent);
        
        // add message body
        parent = document.createElement('div');
        parent.className = 'message';
        
        for (i = 0; i < lines.length; i++)
        {
            words = lines[i].split(' ');
            for (j = 0; j<words.length; j++)
            {
                var word = words[j];
                
                // handle smilies
                if (word in SMILIES)
                {
                    words[j] = document.createElement('img');
                    
                    words[j].src = 'img/smilies/' + SMILIES[word];
                    words[j].alt = word;
                    words[j].title = word;
                    continue;
                }
                
                // handle urls
                if (word.match(url))
                {
                    words[j] = document.createElement('a');
                    words[j].href = word;
                    words[j].target = "_blank";
                    words[j].appendChild(document.createTextNode(word));
                    continue;
                }
                
                // otherwise add to previous string to avoid calling
                // createTextNode too many times
                if (j>0 && (typeof words[j-1] == 'string'))
                {
                     // add to previous word
                    words[j-1] += ' '+words[j];
                    
                    // remove current word
                    words.splice(j,1);
                        
                    // reduce counter since words array has changed
                    j--;
                    continue;
                }
            }
            
            child = document.createElement('p');
            for (j=0; j<words.length; j++)
            {
                if (typeof words[j] == 'string')
                {
                    child.appendChild(
                        document.createTextNode(
                            (j>0?' ':'') + // space before
                            words[j] +
                            (j<words.length-1?' ':'') // space after
                        )
                    );
                }
                else
                {
                    if (j>0 && (typeof words[j-1] != 'string'))
                    {
                        // make sure there is a space between smilies and links
                        child.appendChild(document.createTextNode(' '));
                    }
                    child.appendChild(words[j]);
                }
            }
            
            parent.appendChild(child);
        }
        
        li.appendChild(parent);
        
        /*if (Config.myUsername == msg.sender)
        {
            li.className = "ourself"; 
        }
        */  
        
        return li;
    }
    
    this.createMessageDelimiterHTML = function(date)
    {
        var liEl = document.createElement("li");
        liEl.className = "message-date-delimiter";
        liEl._date = date;
        
        // use day names for the current week and date later
        // TODO: need ot know first day of week. Now we presume Monday
        var firstMondayOfThisWeek = date.clone();
        firstMondayOfThisWeek.moveToDayOfWeek(1, -1);

        if (date.isAfter(firstMondayOfThisWeek))
        {
            liEl.textContent = date.toString("dddd d. MMMM"); // use name of week
        }
        else
        {
            liEl.textContent = date.toString("d. MMMM"); // use date. Monthname
        }
        return liEl;
    }
    
    this.showMessagesForCurrentUser = function()
    {
        var msgs = MessageData.getMessagesForUser(ViewState.CurrentSelectedUser);
        $("message-view-feed").style.display = "none";
        //$("message-view-feed").innerHTML = ""; 
        
        var lastDate = Date.parse("01.01.1970");
        
        // Find last date header or default to Unix epoch time
        for (var i = $('message-view-feed').childNodes.length - 1; i >= 0; i--)
        {
            if ("message-date-delimiter" == $('message-view-feed').childNodes[i].className)
            {
                lastDate = Date.parse($('message-view-feed').childNodes[i]._date);
                break;
            }
        }

        for (var j = 0; j < msgs.length; j++)
        {
            var msg = msgs[j];
            var el = $("message_" + msg.id);
            
            var date = new Date(msg.received);
            date.set(
                {
                    millisecond:0,
                    second:0,
                    minute:0,
                    hour:0
                }
            )
            
            if (el)
            {
                // it exists, only check if state was changed
                var imgEl = el.getElementsByClassName('pendingBox')[0];
                if (msg.state != "pending" && imgEl)
                {
                   imgEl.parentNode.removeChild(imgEl);
                   el.setAttribute('title', '');
                }
            }
            else 
            {
                // detect if there is a new date
                if (! Date.equals(lastDate, date))
                {
                    $("message-view-feed").appendChild(self.createMessageDelimiterHTML(date));
                    lastDate = date;
                }
                
               $("message-view-feed").appendChild(self.createMessageHTML(msg));
               //MessageCtrl.scrollIntoPosition();
            }
        }
        $("message-view-feed").style.display = "block";
        MessageCtrl.scrollIntoPosition();
    }
    
    this.updateMessage = function(post)
    {
        var html = this.createMessageHTML(post);
        if ($('message_' + post.id)) // TODO: should rather do an assert here
        {
            $('message_' + post.id).innerHTML = html.innerHTML;
        }
    }
    
    this.onConversationSelected = function()
    {
        HeaderCtrl.focusConversation();
        MessageCtrl.storeUnsentMessage();
        MessageCtrl.clearAllMessages();
        
        var msgs = MessageData.getConversationsMessages();
        
        for (var j = 0; j < msgs.length; j++)
        {
            var msg = msgs[j];
            $("message-view-feed").appendChild(self.createConversatonHTML(msg));
               
            //MessageCtrl.scrollIntoPosition();
        }
        $("message-view-feed").style.display = "block";

        //MessageCtrl.scrollIntoPosition();
        
        window.location.hash  = "";
    }
    
    this.onSendMessageSubmitted = function(evt)
    {
        var messageText = $("message_text").value;
        messageText = Utility.escape(messageText);
        var body = 'recipient=' + encodeURIComponent(ViewState.CurrentSelectedUser) +
        '&message=' + messageText;
        ViewCtrl.connection.startAndKeep('send', body);
        $("message_text").value = "";
        $("message_text").focus();
        self.onMessageTextInput();
        evt.preventDefault();
        MessageCtrl.clearUnsentMessage();
        return false;
    }

    this.onMessageTextKeyDown = function(evt)
    {
        // send message on enter and newline on shift+enter
        // Will preventdefault on keypress
        if (13 == evt.keyCode && !evt.shiftKey && evt.target.value != "")
        {
            $("send_message").click();
        }
    }
    
    this.onMessageClicked = function(evt)
    {
        var target = evt.target;
        
        // find the li element from which we can fetch the conversation id from
        while (target && 'li' != target.tagName.toLowerCase())
        {
            if (target.parentNode.tagName) // no tag name on HTMLDocument
            {
                target = target.parentNode;
            }
        }
        
        if ('li' == target.tagName.toLowerCase() && target.id && target.id.match(/^conversation_(.+)/))
        {
            var username = $("conversation_" + RegExp.$1).getElementsByTagName("span")[0].textContent;
            ViewCtrl.onUserSelected(username);
        }
    }
    
    this.onMessageTextKeyPress = function(evt)
    {
        if (13 == evt.keyCode && !evt.shiftKey)
        {
            evt.preventDefault();
        }
    }
    
    this.onMessageTextInput = function(evt)
    {
        $("send_message").disabled = !/\S/.test($("message_text").value);
    }
    
    this.init = function()
    {
        $("message-view-feed").addEventListener("scroll", MessageCtrl.onScrollView, false);
        messageViewEl = $("message-view-feed");
    }
}

// TODO: refactor and move methods out into messagectrl and contactsctrl
var ViewCtrl = new function()
{
    var self = this;
    var friendsExpanded = false; // whether or not the friends list is collapsed
    
    this.onContactsClicked = function(evt)
    {
        var targetEl = ("LI" == evt.target.tagName.toUpperCase()) ? evt.target : evt.target.parentNode;
        if (targetEl.id.match(/contact_(.*)/))
        {
            ViewCtrl.onUserSelected(RegExp.$1);
        }
    }
    
    this.onUserSelected = function(username)
    {
        // element references
        var infoEl = null;
        var spanEl = null;
        var imgEl = null;
        
        this.enableForm((username != null && username != ""));
        
        if (username != ViewState.CurrentSelectedUser)
        {
            MessageCtrl.onUserSelected(username);
        }
    }

    this.onFriendRequestsClicked = function()
    {
        ViewState.HideFans = true;
        $("friend-request-container").style.display = "none";
        window.open("http://my.opera.com/" + Config.myUsername + "/friendly/?ref=unite");
    }

    this.onFriendsReceived = function(friends)
    {
        var updatedUsers = [];
        Dbg.inspect(friends, "Friend data=");
        var seenFriends = false;
        
        $("import-friends").style.display = "none";
        
        if (friends["online"])
        {
            seenFriends &= (friends["online"].length > 0);
            for (var i = 0, friend; friend = friends["online"][i]; i++)
            {
                ContactsCtrl.addContact(friend["username"], true, friend);
                updatedUsers.push(friend["username"]);
                
                if (location.hash == '#'+friend.username)
                {
                    ViewCtrl.onUserSelected(friend.username);
                }
            }
        }
        
        if (friends["offline"])
        {
            seenFriends &= (friends["offline"].length > 0);
            for (var i = 0, friend; friend = friends["offline"][i]; i++)
            {
                ContactsCtrl.addContact(friend["username"], false, friend);
                updatedUsers.push(friends["offline"][i]["username"]);
                
                if (location.hash == '#'+friend.username)
                {
                    ViewCtrl.onUserSelected(friend.username);
                }
            }
        }

        if (friends["fans"])
        {
            var changed = (friends["fans"] != ViewState.NumberOfFans);
            if (changed || (ViewState.NumberOfFans > 0 && ! ViewState.HideFans))
            {
                ViewState.NumberOfFans = parseInt(friends["fans"]);
                $("friend-request-container").style.display = "block";
                $("num-fans").textContent = ViewState.NumberOfFans;            
            }
            else
            {
                $("friend-request-container").style.display = "none";
            }
        }
        
        if (friends["removed"])
        {
            for (var i = 0; i < friends["removed"].length; i++)
            {
                this.onUserRemoved(friends["removed"][i].username);
                //updatedUsers.push(friend["username"]);
            }
        }
        
        if (! Config.seenFriends || ! seenFriends)
        {
            $("no-friends-found").style.display = $("find-more-friends").style.display = "block";
        }
        else if (seenFriends && ! Config.seenFriends)
        {
            Config.seenFriends = true;
        }
        
        ViewCtrl.refreshUsers(updatedUsers);
    }
    
    this.onSearchUserNameChanged = function()
    {
        $("search_user_button").disabled = $("add_user_button").disabled = !/\S/.test($("search_username").value);
    }
    
    this.onSearchForUser = function()
    {
        window.open("http://my.opera.com/community/members/search/?q=" + $("search_username").value + "&ref=unite&app=messenger");
    }
    
    this.onAddUser = function()
    {
        // TODO: just for debug
        var recipient = $("search_username").value.split('@');
        var who = recipient[0];
        var where = recipient[1] || null;
    
        ViewCtrl.connection.startAndKeep('friendsadd?who='+encodeURIComponent(who)
            + (where ? '&where=' + encodeURIComponent(where) : ''),
            null);
    }
    
    /**
     * @param userList {string|array}
     */
    this.refreshUsers = function(userList)
    {
        if ("string" == typeof (userList))
        {
            userList = [userList]; 
        }
        //Dbg.inspect(userList, "Refresh:");

        for (var i = 0; i < userList.length; i++)
        {
            if (ViewState.CurrentSelectedUser == userList[i])
            {
                MessageCtrl.showMessagesForCurrentUser();
                ContactsCtrl.updateState(userList[i]);
                //var body = 'user=' + encodeURIComponent(ViewState.CurrentSelectedUser);
                //ViewCtrl.connection.startAndKeep('readmessages', body);
            }
            else
            {
                //opera.postError("refresh " + userList[i]);
                ContactsCtrl.updateState(userList[i]);
            }
        }
        
        self.updateFriendsMoreButton();
        
    }
    
    /**
     * Starts adding messages asynchronously.
     *
     * If called again the previous one will stop.
     *
     * To stop it alltogether send an empty array.
     *
     * @param messages {Array} An array of messages
     */
    
    this.addMessagesAsync = new function()
    {
        var MESSAGES_PER_ITERATION = 1;
        
        var interval = 0;
        var msgs = null;
        var index = 0;
        
        /** below is the actual function that gets called **/
        
        return function(messages)
        {
            msgs = messages;
            index = msgs.length-1;
            
            if (interval)
            {
                clearInterval(interval); // stop any existing interval
            }

            addMoreMessages();
            interval = setInterval(addMoreMessages, 50); // give some time for the browser to breath
        }
        
        function addMoreMessages()
        {
            var msgParts = null;
            
            // dom variables
            var feed = $("message-view-feed");
            var frag = document.createDocumentFragment();            
            var ele = null; /* generic temporary element */
            
            var i;
            var j;
            
            var scrollHeightStart = feed.scrollHeight;
            
            for (i = Math.max(0, index-MESSAGES_PER_ITERATION+1); i <= index; i++)
            {
                msg = msgs[i];
                ele = $("message_" + msg.id);
                
                if (ele)
                {
                    // TODO: probably state change from pending to sent
                    ele = ele.querySelector('label img');
                    if (ele && msg.state != "pending")
                    {
                       ele.parentNode.removeChild(ele);
                    }
                }
                else
                {
                    //frag.appendChild(self.createMessageHTML(msg));
                }
            }
            index -= MESSAGES_PER_ITERATION;
            
            if (index < 0)
            {
                clearInterval(interval);
                interval = 0;
            }
            
            // insert the fragment at the top of the list
            if (feed.firstChild)
            {
                feed.insertBefore(frag, feed.firstChild);
            }
            else
            {
                feed.appendChild(frag);
            }
            
            feed.scrollTop += feed.scrollHeight - scrollHeightStart;
            
            MessageCtrl.scrollIntoPosition();
        }
    };
    
    this.onUserRemoved = function(username)
    {
        ContactsCtrl.removeContact(username);
        ViewCtrl.onUserSelected();
    }
    
    this.debugRemoveUser = function()
    {
        var body =  'who='+ViewState.CurrentSelectedUser;
        ViewCtrl.connection.startAndKeep('friendsremove?'+body, null);
    }
    
    this.debugRemoveFan = function()
    {
        ViewCtrl.onFriendsReceived( { fansNoMore : [ViewState.CurrentSelectedUser] } );
    }
    
    this.clearAllMessages = function()
    {
        ViewCtrl.connection.startAndKeep('clear', null);
    }
    
    this.displayAlertScreen = function()
    {
        $("alert_dialog").style.display = "block";
    }
    
    this.hideAlertScreen = function()
    {
        $("alert_dialog").style.display = "none";
    }
    
    /**
     * Shows the intro screen for users who have no friends.
     */
    
    this.displayEmptyScreen = function()
    {
        $('app').removeClass('loading');
        $('app').addClass('empty');
        self.enableForm(false);
    };
    
    /**
     * Shows the loading screen
     */

    this.displayLoadingScreen = function()
    {
        $('app').removeClass('empty');
        $('app').addClass('loading');
        self.enableForm(false);
    };
    
    /**
     * Shows the normal screen
     */

    this.displayNormalScreen = function()
    {
        $('app').removeClass('loading');
        $('app').removeClass('empty');
        
        if (ViewState.CurrentSelectedUser)
        {
            self.enableForm(true); // enable if a user is selected
        }
    };
    
    this.displayNoUsersScreen = function()
    {
        $("import-friends").style.display = "none";
        $("no-friends-found").style.display = $("find-more-friends").style.display = "block";
    }
    
    /**
     * Handles disabling form elements and hiding smilies etc.
     *
     * @param enable {Boolean} Whether or not the form should be enabled.
     */
    
    this.enableForm = function(enable, noHelp)
    {
        $("send_message").disabled = !enable || !/\S/.test($("message_text").value);
        $("message_text").disabled = !enable;
        $('smilies_expand').disabled = !enable;
        $("message-filter").style.display = enable ? 'block' : 'none';
        $("message-view-feed").style.display = enable ? 'block' : 'none';
        $("start-help").style.display = (enable || noHelp) ? 'none' : 'block';
        $("help_header").style.display = (enable || noHelp)? 'block' : 'none';
        if (!enable)
        {
            self.showSmilies(false);
        }
        
        if ($("friends-offline").childNodes.length > 0)
        {
            $("friends-offline-header").style.display = "block";
        }
        else
        {
            $("friends-offline-header").style.display = "none";
        }
    }
    
    /**
     * Shows/hides the smilies
     *
     * @param show {Boolean} If true then show them, otherwise hide them
     */
    
    this.showSmilies = function(show)
    {
        var parent = document.getElementById('message_smilies');
        var smilies = parent.getElementsByTagName('ul')[0].getElementsByTagName('button');
        
        if (show)
        {
            parent.addClass('expanded');
        }
        else
        {
            
            parent.removeClass('expanded');
        }
       // alert(show);
        // important to disable them in case they were focused when hidden
        // otherwise the user can press enter and their 'click' event will call
        for (var i=0; i<smilies.length; i++)
        {
            smilies[i].disabled = !show;
        }
    };
    
    this.addSmileyEvents = function()
    {
        var parent = $('message_smilies');
        var show = $('smilies_expand');
        var list = parent.getElementsByTagName('ul')[0];
        var smilies = list.getElementsByTagName('button');
        
        show.addEventListener('click', function(evt)
        {
            if (!parent.hasClass('expanded'))
            {
                self.showSmilies(true);
                evt.stopPropagation(); // make sure collapse event handler isnt called
            }
        }, true);
        
        document.addEventListener('click', function(evt)
        {
            if (parent.hasClass('expanded'))
            {
                self.showSmilies(false);
            }
        }, false);
        
        // just have 1 event listener on the list, spatnav will still focus
        // the smilies since they are button elements
        list.addEventListener('click', function(evt)
        {
            var target = evt.target.getElementsByTagName('img')[0];
            var img = null;
            var textarea = $('message_text');
            
            // may have clicked the image element inside the button
            if (target.parentNode instanceof window.HTMLButtonElement)
            {
                target = target.parentNode;
            }
            if (!target.parentNode instanceof window.HTMLButtonElement)
            {
                return; // somewhere in the list other than the button was clicked
            }
            
            // phew, we now have the button that was click
            
            img = target.getElementsByTagName('img')[0];
            
            textarea.value += (/(\s|^)$/.test(textarea.value) ? '' : ' ') + // space before smiley only if necessary
                              img.getAttribute('alt') + // the smiley code
                              ' '; // space after smiley
            
            textarea.focus();            
            MessageCtrl.onMessageTextInput.call(textarea, evt); // make sure any event relating to change of value gets picked up
            
        }, true);
    }
    
    this.friendsMoreClick = function(evt)
    {
        friendsExpanded = !friendsExpanded;
        
        if (friendsExpanded)
        {
            $('contacts').addClass('expanded');
            $('friends-more').innerHTML = _('Collapse');
        }
        else
        {
            $('contacts').removeClass('expanded');
            $('friends-more').innerHTML = _('Show All');
        }
    };
    
    this.updateFriendsMoreButton = function()
    {
        /* hack to keep a max-width without causing reflow problems */
        /* TODO: solve this hopefully with CSS */
        
        var SCALE_WIDTH = 92; // the percent width when space is less than MAX_WIDTH
        var MAX_WIDTH = 960; // the maximum width to let the 
        var width = document.body.offsetWidth;       
        
        /* this will keep 'app' at the width defined by SCALE_WIDTH and MAX_WIDTH */
        $('app').style.width = width < MAX_WIDTH*100/SCALE_WIDTH ? SCALE_WIDTH+'%' : Math.round(MAX_WIDTH*100*100/width)/100 + '%';
        $('app').style.maxWidth = '';
        
        // now do the showing/hiding of the friendsMore button...
        
        // remove the class then perhaps add it again,
        // this allows just one logic to solve whether or not
        // we have contact list overflow
        $('contacts').removeClass('overflowing');
        
        if (
            $('contacts').offsetHeight <=
                $('contacts-friends').offsetHeight
              + $('add-friends').offsetHeight
              - $('friends-more').offsetHeight
        )
        {
            $('contacts').addClass('overflowing');
        }
        
        if ($('friends-online').childNodes.length == 0)
        {
            $('no-online-friends').style.display = 'block';
            $('friends-online-header').removeClass("empty");
        }
        else
        {
            $('no-online-friends').style.display = 'none';
            $('friends-online-header').addClass("empty");
        }
        
        if ($('friends-offline').childNodes.length == 0)
        {
            $('no-offline-friends').style.display = 'block';
            $('friends-offline-header').removeClass("empty");
        }
        else
        {
            $('no-offline-friends').style.display = 'none';
            $('friends-offline-header').addClass("empty");
        }
    };
    
    this.debugAddUnread = function()
    {
        if (ViewState.CurrentSelectedUser)
        {
            MessageData.increaseNumUnread(ViewState.CurrentSelectedUser);
        }
        
        ViewCtrl.refreshUsers(ViewState.CurrentSelectedUser);
    }
    
    this.clearUser = function()
    {
        ViewCtrl.onUserSelected("");
    }
    
    this.init = function(connection)
    {
        this.connection = connection;
        
        $("search_user_button").addEventListener("click", ViewCtrl.onSearchForUser, false);
        $("friends-more").addEventListener("click", ViewCtrl.friendsMoreClick, false);
        $("add_user_button").addEventListener("click", ViewCtrl.onAddUser, false);
        $("search_username").addEventListener("keyup", ViewCtrl.onSearchUserNameChanged, false);
        $("send_form").addEventListener("submit", MessageCtrl.onSendMessageSubmitted, false);
        $("message_text").addEventListener("keydown", MessageCtrl.onMessageTextKeyDown, false);
        $("message_text").addEventListener("keypress", MessageCtrl.onMessageTextKeyPress, false);
        $("message_text").addEventListener("input", MessageCtrl.onMessageTextInput, false);
        $("help_header").addEventListener("click", MessageCtrl.onHelpClicked, false);
        $("username_header").addEventListener("click", HeaderCtrl.onUserClicked, false);
        //$("conversation_header").addEventListener("click", HeaderCtrl.onConversationClicked, false);
        $("message-view-feed").addEventListener("click", MessageCtrl.onMessageClicked, false);
        $("message-filter").addEventListener("click", MessageFilterCtrl.onFilterClicked, false);
        $("friend-requests").addEventListener("click", ViewCtrl.onFriendRequestsClicked, false);
        
        this.addSmileyEvents();
        
        /** debug **/
        $("debug_empty").addEventListener("click", self.displayEmptyScreen, false);
        $("debug_loading").addEventListener("click", self.displayLoadingScreen, false);
        $("debug_normal").addEventListener("click", self.displayNormalScreen, false);
        $("debug_clear_user").addEventListener("click", self.clearUser, false);
        $("debug_bring_offline").addEventListener("click", ViewCtrl.displayAlertScreen, false);
        
        //this.onUserSelected(null);
        this.onSearchUserNameChanged();
        
        // TODO:debug
        $("remove_user").addEventListener("click", ViewCtrl.debugRemoveUser, false);
        $("remove_fan").addEventListener("click", ViewCtrl.debugRemoveFan, false);
        $("clear_all_messages").addEventListener("click", ViewCtrl.clearAllMessages, false);
        
        window.addEventListener('resize', self.updateFriendsMoreButton.wait(), false);
        self.updateFriendsMoreButton();
        
        MessageCtrl.init();
    }
}
