
var MessageFilterSpan =
{
    All : 0,
    Today : 1,
    Yesterday : 2,
    LastWeek : 3,
    LastMonth : 4
}

MessageFilterSpan["toString"] = function(filterId)
{
    for (var i in MessageFilterSpan)
    {
        if (filterId === MessageFilterSpan[i])
        {
            return i;
        }
    }
    
    return "";
}

var MessageFilter = new function()
{
    var filter = {};
    
    this.setFilterForContact = function(username, filter)
    {
        filter[username] = filter; 
    }
    
    this.getFilterForContact = function(username)
    {
        return filter[username];
    }

    this.getFromDate = function(span)
    {
        var now = new Date();
        
        if (! span)
        {
            span = 0;
        }
        
        switch (span.toString())
        {
            case "1":
            {
                now = new Date();
                break;
            }
            case "2":
            {
                now.setTime(now.getTime()-86400000);
                break;
            }
            case "3":
            {
                now.setTime(now.getTime()-604800000);
                break;
            }
            case "4":
            {
                now.setTime(now.getTime()-2419200000);
                break;
            }
            default:
            {
                now = new Date(0); 
            }
        }
        
        if (span != 0)
        {
            now.setMilliseconds(0);
            now.setSeconds(0);
            now.setMinutes(0);
            now.setHours(0);
        }

        return now.getTime();
    }
    
    /* returns the last filter that is applicable to the time */
    this.getLastFilter = function(time)
    {
        if (! time || time > this.getFromDate(MessageFilterSpan.Today))
        {
            return MessageFilterSpan.Today;
        }
        else if (time > this.getFromDate(MessageFilterSpan.Yesterday))
        {
            return MessageFilterSpan.Yesterday;
        }
        else if (time > this.getFromDate(MessageFilterSpan.LastWeek))
        {
            return MessageFilterSpan.LastWeek;
        }
        else if (time > this.getFromDate(MessageFilterSpan.LastMonth))
        {
            return MessageFilterSpan.LastMonth;
        }
        
        return MessageFilterSpan.All;
    }
}

var MessageData = new function()
{
    var data = {};                  // assoc array, username is key. Value is an assoc array containing the id
    var sortedData = {};            // contians message ids sorted by time
    var idToMsg = {};               // assoc array, id to message
    var isSorted = {};              // assoc array keeping track of which messagearrays are sorted
    var numUnread = {};             // number of unread for user
    var numPending = {};            // number of pending messages
    var hadPendingOnStartup = {};   // store pending state on startup.
    var isOnlineState = {};         // is online?
    var isFan = {};                 // is the user a fan?
    var firstMessageTime = {};      // stores the time of the first message
    var lastFilter = {};            // caches the "last" filter allowed based on firstMessageTime
    var messageFilter = {};         // stores the currently selected messagefilter in the ui
    
    this.deletePendingMessage = function(post)
    {
        data[post.recipient] = sortedData[post.recipient] = [];
        data[post.recipient][post.id] = sortedData[post.recipient][post.id] = null;
        delete data[post.recipient][post.id];
        delete sortedData[post.recipient][post.id];
        numPending[post.recipient]--;
    }
    
    this.addMessage = function(post)
    {
        var isDirty = false;
        var imTheSender = (Config.myUsername == post.sender);
        var conversationName = imTheSender ? post.recipient : post.sender; // determine which array to store it in. Will store "my mesages" in "their" store
        if (! data[conversationName])
        {
            data[conversationName] = [];
        }
        if (! sortedData[conversationName])
        {
            sortedData[conversationName] = [];
        }
        
        // Unescape body
        post.message = Utility.unescape(post.message);
        
        if (data[conversationName][post.id])
        {
            if ("pending" == data[conversationName][post.id].state && "pending" !== post.state)
            {
                hadPendingOnStartup[conversationName] = false; // don't need this anymore
                numPending[conversationName]--;
                
                // also need to look through sortedData. This is really an ugly hack and the
                // storage of the mesages in the client should be refactored. 
                for (var i = 0; i < sortedData[conversationName].length; i++)
                {
                    if (sortedData[conversationName][i].id == post.id)
                    {
                        sortedData[conversationName][i] = post;
                        break;
                    }
                }
                
            }
            if (data[conversationName][post.id].state != post.state)
            {
                isDirty = true;
            }
            
            data[conversationName][post.id] = post;
        }
        else
        {
            data[conversationName][post.id] = post;

            if ("unread" == post.state &&
                Config.myUserName != post.sender &&
                ViewState.CurrentSelectedUser != conversationName)
            {
                this.increaseNumUnread(conversationName);
            }
            else if ("pending" == post.state)
            {
                hadPendingOnStartup[conversationName] = false; // don't need this anymore
                
                if (! numPending[conversationName])
                {
                    numPending[conversationName] = 1;
                }
                else
                {
                    numPending[conversationName]++;
                }
            }
            
            sortedData[conversationName].push(post);
        }

        return isDirty;
    }
    
    this.getMessagesForUser = function(username)
    {
        // ensure they are sorted
        if (! isSorted[username])
        {
            MessageData.sortByTime(username);
        }
        
        var filter = MessageData.getMessageFilter(username);
        var fromDate = 0;
        var returnData = [];
        var resultNum = 0;
        var debugNumPass = 0;
        var userData = sortedData[username];
        
        // TODO: need better search
        if (userData && userData.length > 0)
        {
            fromDate = MessageFilter.getFromDate(filter);
            for (var i = userData.length-1; i >= 0; i--)
            {
                if (userData[i].received > fromDate)
                {
                    returnData.unshift(userData[i]);
                }
                else
                {
                    break;
                }
            }
        }
        
        return returnData;

      /* failed try on interpolation search 
         if (userData && userData.length > 0)
        {
            var low = 0, mid, high = userData.length-1;
            fromDate = MessageFilter.getFromDate(filter);
            opera.postError("-->" + userData[low].received + " " + userData[high].received + " " + fromDate + " " + (userData[low].received < fromDate) +
                          " " + (userData[high].received >= fromDate));
            while (userData[low].received < fromDate && userData[high].received >= fromDate) // interpolatation search 
            {
                debugNumPass++;
                mid = low + Math.round(((fromDate - userData[low].received) * (high - low)) / (userData[high].received - userData[low].received));
                opera.postError("Mid is now " + mid + " L=" + low + " H=" + high);
                if (userData[mid].received < fromDate)
                {
                    low = mid + 1;
                }
                else if (userData[mid].received > fromDate)
                {
                    high = mid - 1;
                }
                else
                {
                    resultNum = mid;
                    opera.postError("found " + mid);
                    break;
                }
                
                opera.postError("-->" + userData[low].received + " " + userData[high].received + " " + fromDate + " " + (userData[low].received < fromDate) +
                          " " + (userData[high].received >= fromDate));
            }
            
            if (userData[low].received == fromDate)
            {
                resultNum = low;
            }
            else 
            {
                resultNum = userData.length;
            }
            
            
            returnData = userData.slice(resultNum);
            
        }*/


    }
    
    this.getConversationsMessages = function()
    {
        var msgs = [];
        
        for (var username in sortedData)
        {
            if (! isSorted[username])
            {
                MessageData.sortByTime(username);
            }
            
            if (sortedData[username].length)
            {
                Dbg.inspect(sortedData, "Sorted for " + username + ":");
                var msg = sortedData[username][sortedData[username].length-1];
                msgs.push(msg);
            }
        }
        Dbg.inspect(msgs, "ConversationsStored:");
        var bytime = function(a, b)
        {
            return +a.received - +b.received;
        }
        
        return msgs.sort(bytime);
    }
    
    this.hasMessages = function(username)
    {
        return (data[username] && data[username].length > 0);
    }
    
    this.increaseNumUnread = function(username)
    {
        if (! numUnread[username])
        {
            numUnread[username] = 0;
        }
        numUnread[username]++;
    }
    
    this.setNumUnread = function(username, num)
    {
        numUnread[username] = num;
    }
    
    this.getNumUnread = function(username)
    {
        return numUnread[username];
    }
    
    this.hasPending = function(username)
    {
        if (hadPendingOnStartup[username])
        {
            return true;
        }
        else
        {
            return ((numPending[username] > 0) || false);
        }
    }
    
    this.setPending = function(username, pending)
    {
        //numPending[username] = pending ? 1 : 0;
        hadPendingOnStartup[username] = pending;
    }
    
    this.setIsOnline = function(username, isOnline)
    {
        isOnlineState[username] = isOnline;
    }
    
    this.isOnline = function(username)
    {
        return isOnlineState[username];
    }
    
    this.setIsFan = function(username, isfan)
    {
        isFan[username] = isfan;
    }
    
    this.isFan = function(username)
    {
        return isFan[username];
    }
    
    this.getFirstMessageTime = function(username)
    {
        return firstMessageTime[username];
    }
    
    this.setFirstMessageTime = function(username, time)
    {
        var filter = MessageFilter.getLastFilter(time);
        lastFilter[username] = filter;
        firstMessageTime[username] = time;
    }
    
    this.getLastFilter = function(username)
    {
        return lastFilter[username];
    }
    
    this.getMessageFilter = function(username)
    {
        return messageFilter[username] || MessageFilterSpan.Today;
    }
    
    this.setMessageFilter = function(username, filter)
    {
        messageFilter[username] = filter;
    }
    
    this.sortByTime = function(username)
    {
        var bytime = function(a, b)
        {
            return +a.received - +b.received;
        }
        if (data[username])
        {
            isSorted[username] = true;
            sortedData[username] = sortedData[username].sort(bytime);   
        }
    }
}
