/**
 * Call to return a wrapper for the function which
 * ensures the function is not called too often.
 *
 * Used in the window resize event handler to avoid flooding of calls.
 */

Function.prototype.wait = function()
{
    var self = this;
    var timeout = 0;
    
    function callFunct()
    {
        timeout = 0;
        self();
    }
    
    return function()
    {
        if (!timeout)
        {
            timeout = setTimeout(callFunct, 50);
        }
    };
}

var $ = function(elId) { return document.getElementById(elId); }

var ViewState =
{
    CurrentSelectedUser : null,
    IsFriendsPaneExpanded : false, // is the friendspane expanded
    HideFans : false,
    NumberOfFans : 0,
    save : function()
    {
        // TODO: save state on server
    },
    load : function()
    {
        // TODO: load state from server
    },
    seenFriends : false
}

var Utility =
{
    escape: function(str)
    {
        str = str.replace(/\\/g,'\\\\'); // turn aaa\bbb into aaa\\bbb
        str = str.replace(/(\r?\n)|(\r)/g,'\\n'); // turn aaa[newline]bbb into aaa\nbbb
        str = str.replace(/\&/g, "%26");
        str = str.replace(/\+/g, "%2B");
        str = str.replace(/\%/g, "%25");
        return str;
    },
    
    unescape: function(str)
    {
        str = str.replace(/%26/g, "&");
        str = str.replace(/%2B/g, "+");
        str = str.replace(/%25/g, "%");
        str = str.replace(/([^\\](\\\\)*)\\n/g,'$1\n'); // introduce new-lines
        str = str.replace(/([^\\](\\\\)*)\\n/g,'$1\n'); // do it again since the previous doesn't catch all
        str = str.replace(/\\\\/g,'\\'); // collapse double backslash
        return str;
    }
}

var Dbg =
{
    inspect : function(o, prefix)
    {
        //opera.postError("[" + (new Date()).toString("HH:MM:ss") + "]: " + prefix + "\n" + JSON.stringify(o));
    }
}

if (! HTMLElement.prototype.hasClass)
{
    Element.prototype.hasClass = function(className)
    {
      return (' '+this.className+' ').indexOf(' '+className+' ') >= 0;
    }
}

if (! HTMLElement.prototype.addClass)
{
    HTMLElement.prototype.addClass = function(className)
    {
        if (! this.hasClass(className))
        {
            this.className += " " + className;
        }
    }
}
    
if (! HTMLElement.prototype.removeClass)
{
    HTMLElement.prototype.removeClass = function(className)
    {
        this.className = this.className.replace(className, "");
    }
}
